# -*- coding: utf8 -*-

import os
import numpy as np
import tensorflow as tf

from yarll.agents.agent import Agent
from yarll.memory.memory import Memory
from yarll.misc.noise import OrnsteinUhlenbeckActionNoise
from yarll.misc.network_ops import batch_norm_layer, fan_in_initializer, linear_fan_in

class DDPG(Agent):
    def __init__(self, env, monitor_path: str, **usercfg) -> None:
        super(DDPG, self).__init__(**usercfg)
        self.env = env
        self.monitor_path: str = monitor_path

        self.config.update(
            n_episodes=100000,
            n_timesteps=env.spec.tags.get("wrapper_config.TimeLimit.max_episode_steps"),
            actor_learning_rate=1e-4,
            critic_learning_rate=1e-3,
            ou_theta=0.15,
            ou_sigma=0.2,
            gamma=0.99,
            batch_size=64,
            tau=0.001,
            l2_loss_coef=1e-2,
            n_actor_layers=2,
            n_hidden_units=64,
            actor_layer_norm=True,
            critic_layer_norm=False,  # Batch norm for critic does not seem to work
            replay_buffer_size=1e6,
            replay_start_size=10000  # Required number of replay buffer entries to start training
        )
        self.config.update(usercfg)

        self.state_shape: list = list(env.observation_space.shape)
        self.n_actions: int = env.action_space.shape[0]
        self.states = tf.placeholder(tf.float32, [None] + self.state_shape, name="states")
        self.actions_taken = tf.placeholder(tf.float32, [None, self.n_actions], name="actions_taken")
        self.critic_target = tf.placeholder(tf.float32, [None, 1], name="critic_target")
        self.is_training = tf.placeholder(tf.bool, name="is_training")

        with tf.variable_scope("actor"):
            self.action_output, self.actor_vars = self.build_actor_network()

        self.target_action_output, actor_target_update = self.build_target_actor_network(self.actor_vars)

        self.q_gradient_input = tf.placeholder("float", [None, self.n_actions], name="q_grad_input")
        self.actor_policy_gradients = tf.gradients(
		          self.action_output, self.actor_vars, -self.q_gradient_input, name="actor_gradients")
        self.actor_train_op = tf.train.AdamOptimizer(
            self.config["actor_learning_rate"],
            name="actor_optimizer").apply_gradients(list(zip(self.actor_policy_gradients, self.actor_vars)))

        with tf.variable_scope("critic"):
            self.q_value_output, self.critic_vars = self.build_critic_network()

        self.target_q_value_output, critic_target_update = self.build_target_critic_network(self.critic_vars)

        l2_loss = tf.add_n([self.config["l2_loss_coef"] * tf.nn.l2_loss(var) for var in self.critic_vars])
        self.critic_loss = tf.reduce_mean(tf.square(self.critic_target - self.q_value_output)) + l2_loss
        self.critic_train_op = tf.train.AdamOptimizer(
            self.config["critic_learning_rate"],
            name="critic_optimizer").minimize(self.critic_loss)
        self.action_gradients = tf.gradients(self.q_value_output, self.actions_taken, name="action_gradients")

        summaries = []
        for v in self.actor_vars + self.critic_vars:
            summaries.append(tf.summary.histogram(v.name, v))
        self.model_summary_op = tf.summary.merge(summaries)

        self.update_targets_op = tf.group(actor_target_update, critic_target_update, name="update_targets")


        self.action_noise = OrnsteinUhlenbeckActionNoise(
            self.n_actions,
            self.config["ou_sigma"],
            self.config["ou_theta"]
        )

        self.replay_buffer = Memory(int(self.config["replay_buffer_size"]))

        self.session = tf.Session()
        self.init_op = tf.global_variables_initializer()

        self.n_updates = 0

        self.summary_writer = tf.summary.FileWriter(os.path.join(
            self.monitor_path, "summaries"), tf.get_default_graph())

    def _initalize(self):
        self.session.run(self.init_op)

    def build_actor_network(self):
        layer1_size = 400
        layer2_size = 300

        x = self.states
        if self.config["actor_layer_norm"]:
            x = batch_norm_layer(x, training_phase=self.is_training, scope_bn="batch_norm_0", activation=tf.identity)
        with tf.variable_scope("L1"):
            x, l1_vars = linear_fan_in(x, layer1_size)
            if self.config["actor_layer_norm"]:
                x = batch_norm_layer(x, training_phase=self.is_training, scope_bn="batch_norm_1", activation=tf.nn.relu)
        with tf.variable_scope("L2"):
            x, l2_vars = linear_fan_in(x, layer2_size)
            if self.config["actor_layer_norm"]:
                x = batch_norm_layer(x, training_phase=self.is_training, scope_bn="batch_norm_2", activation=tf.nn.relu)

        with tf.variable_scope("L3"):
            W3 = tf.Variable(tf.random_uniform([layer2_size, self.n_actions], -3e-3, 3e-3), name="w")
            b3 = tf.Variable(tf.random_uniform([self.n_actions], -3e-3, 3e-3), name="b")
            action_output = tf.tanh(tf.nn.xw_plus_b(x, W3, b3))
            l3_vars = [W3, b3]

        return action_output, l1_vars + l2_vars + l3_vars

    def build_target_actor_network(self, actor_vars: list):
        ema = tf.train.ExponentialMovingAverage(decay=1 - self.config["tau"])
        target_update = ema.apply(actor_vars)
        target_net = [ema.average(v) for v in actor_vars]

        x = self.states
        if self.config["actor_layer_norm"]:
            x = batch_norm_layer(
                x, training_phase=self.is_training, scope_bn="target_batch_norm_0", activation=tf.identity)

        x = tf.nn.xw_plus_b(x, target_net[0], target_net[1])
        if self.config["actor_layer_norm"]:
            x = batch_norm_layer(
                x, training_phase=self.is_training, scope_bn="target_batch_norm_1", activation=tf.nn.relu)
        x = tf.nn.xw_plus_b(x, target_net[2], target_net[3])
        if self.config["actor_layer_norm"]:
            x = batch_norm_layer(
                x, training_phase=self.is_training, scope_bn="target_batch_norm_2", activation=tf.nn.relu)

        action_output = tf.tanh(tf.nn.xw_plus_b(x, target_net[4], target_net[5]))

        return action_output, target_update

    def build_critic_network(self):
        layer1_size = 400
        layer2_size = 300

        x = self.states
        with tf.variable_scope("L1"):
            if self.config["critic_layer_norm"]:  # Defaults to False (= don't use it)
                x = batch_norm_layer(x, training_phase=self.is_training,
                                     scope_bn="batch_norm_0", activation=tf.identity)
            x, l1_vars = linear_fan_in(x, layer1_size)
            x = tf.nn.relu(x)
        with tf.variable_scope("L2"):
            W2 = tf.get_variable(
                "w", [layer1_size, layer2_size], initializer=fan_in_initializer(layer1_size + self.n_actions))
            W2_action = tf.get_variable(
                "w_action", [self.n_actions, layer2_size], initializer=fan_in_initializer(layer1_size + self.n_actions))
            b2 = tf.get_variable(
                "b", [layer2_size], initializer=fan_in_initializer(layer1_size + self.n_actions))
            x = tf.nn.relu(tf.matmul(x, W2) + tf.matmul(self.actions_taken, W2_action) + b2)
        with tf.variable_scope("L3"):
            W3 = tf.Variable(tf.random_uniform([layer2_size, 1], -3e-3, 3e-3), name="w")
            b3 = tf.Variable(tf.random_uniform([1], -3e-3, 3e-3), name="b")
            q_value_output = tf.nn.xw_plus_b(x, W3, b3, name="q_value")

        return q_value_output, l1_vars + [W2, W2_action, b2, W3, b3]

    def build_target_critic_network(self, critic_vars: list):

        ema = tf.train.ExponentialMovingAverage(decay=1 - self.config["tau"])
        target_update = ema.apply(critic_vars)
        target_net = [ema.average(v) for v in critic_vars]

        x = self.states
        if self.config["critic_layer_norm"]:
            x = batch_norm_layer(x, training_phase=self.is_training, scope_bn="batch_norm_0", activation=tf.identity)
        x = tf.nn.relu(tf.nn.xw_plus_b(x, target_net[0], target_net[1]))
        x = tf.nn.relu(tf.matmul(x, target_net[2]) + tf.matmul(self.actions_taken, target_net[3]) + target_net[4])
        q_value_output = tf.nn.xw_plus_b(x, target_net[5], target_net[6])

        return q_value_output, target_update

    def actor_gradients(self, state_batch: np.ndarray, action_batch: np.ndarray):
        q, grads = tf.get_default_session().run([self.q_value_output, self.action_gradients], feed_dict={
            self.states: state_batch,
            self.actions_taken: action_batch,
            self.is_training: False
        })
        summary = tf.Summary()
        summary.value.add(tag="model/actor_loss", simple_value=float(-np.mean(q)))
        self.summary_writer.add_summary(summary, self.n_updates)
        return grads[0]

    def target_q(self, states: np.ndarray, actions: np.ndarray):
        return tf.get_default_session().run(self.target_q_value_output, feed_dict={
            self.states: states,
            self.actions_taken: actions,
            self.is_training: False
        })

    def q_value(self, states: np.ndarray, actions: np.ndarray):
        return tf.get_default_session().run(self.q_value_output, feed_dict={
            self.states: states,
            self.actions_taken: actions,
            self.is_training: False
            })

    def actions(self, states: np.ndarray) -> np.ndarray:
        """Get the actions for a batch of states."""
        return tf.get_default_session().run(self.action_output, feed_dict={
            self.states: states,
            self.is_training: True
        })

    def action(self, state: np.ndarray) -> np.ndarray:
        """Get the action for a single state."""
        return tf.get_default_session().run(self.action_output, feed_dict={
            self.states: [state],
            self.is_training: False
        })[0]

    def target_actions(self, states: np.ndarray) -> np.ndarray:
        """Get the actions for a batch of states using the target actor network."""
        return tf.get_default_session().run(self.target_action_output, feed_dict={
            self.states: states,
            self.is_training: True
        })

    def train(self):
        sample = self.replay_buffer.get_batch(int(self.config["batch_size"]))

        # for n_actions = 1
        action_batch = np.resize(sample["actions"], [int(self.config["batch_size"]), self.n_actions])

        # Calculate critic targets
        next_action_batch = self.target_actions(sample["states1"])
        q_value_batch = self.target_q(sample["states1"], next_action_batch)
        critic_targets = sample["rewards"] + (1 - sample["terminals1"]) * \
            self.config["gamma"] * q_value_batch.squeeze()
        critic_targets = np.resize(critic_targets, [int(self.config["batch_size"]), 1]).astype(np.float32)
        # Update actor weights
        fetches = [self.q_value_output, self.critic_loss, self.critic_train_op]
        predicted_q, critic_loss, _ = tf.get_default_session().run(fetches, feed_dict={
            self.critic_target: critic_targets,
            self.states: sample["states0"],
            self.actions_taken: action_batch,
            self.is_training: True
        })

        summary = tf.Summary()
        summary.value.add(tag="model/critic_loss", simple_value=float(critic_loss))
        summary.value.add(tag="model/predicted_q_mean", simple_value=np.mean(predicted_q))
        summary.value.add(tag="model/predicted_q_std", simple_value=np.std(predicted_q))
        self.summary_writer.add_summary(summary, self.n_updates)

        # Update the actor using the sampled gradient:
        action_batch_for_gradients = self.actions(sample["states0"])
        q_gradient_batch = self.actor_gradients(sample["states0"], action_batch_for_gradients)

        tf.get_default_session().run(self.actor_train_op, feed_dict={
            self.q_gradient_input: q_gradient_batch,
            self.states: sample["states0"],
            self.is_training: True
        })

        # Update the target networks
        tf.get_default_session().run([self.update_targets_op, self.model_summary_op])
        self.n_updates += 1

    def noise_action(self, state: np.ndarray):
        """Choose an action based on the actor and exploration noise."""
        action = self.action(state)
        return action + self.action_noise()

    def learn(self):
        max_action = self.env.action_space.high
        self._initalize()
        with self.session as sess, sess.as_default():
            for episode in range(int(self.config["n_episodes"])):
                state = self.env.reset()
                episode_reward = 0
                episode_length = 0
                for _ in range(int(self.config["n_timesteps"])):
                    action = self.noise_action(state)
                    new_state, reward, done, _ = self.env.step(action * max_action)
                    episode_length += 1
                    episode_reward += reward
                    self.replay_buffer.add(state, action, reward, new_state, done)
                    if self.replay_buffer.n_entries > self.config["replay_start_size"]:
                        self.train()
                    state = new_state
                    if done:
                        self.action_noise.reset()
                        summary = tf.Summary()
                        summary.value.add(tag="global/Episode_length",
                                          simple_value=float(episode_length))
                        summary.value.add(tag="global/Reward",
                                          simple_value=float(episode_reward))
                        self.summary_writer.add_summary(summary, episode)
                        self.summary_writer.flush()
                        break



# -*- coding: utf8 -*-

class WrongArgumentsError(Exception):
    """
    The program was called with incorrect arguments or an incorrect combination of them.
    """
    pass

class WrongShapeError(Exception):
    """
    A sequence has the wrong shape.
    """
    pass

class ClassNotRegisteredError(Exception):
    """
    Tried to create an environment or agent instance that is not registered.
    """
    pass



# -*- coding: utf8 -*-

import numpy as np

class Sarsa(object):
    """Sarsa learner for function approximation"""
    def __init__(self, gamma: float, alpha: float, policy, traces, function_approximation, actions, start_state, start_action) -> None:
        super(Sarsa, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        self.policy = policy
        self.traces = traces
        self.function_approximation = function_approximation
        self.actions = actions
        self.old_state = start_state
        self.old_action = start_action

    def step(self, state: np.ndarray, reward: float):
        """Do one step of updating traces, function approximation and selecting an action using a policy"""
        self.traces.replacing_traces(self.function_approximation.present_features(self.old_state, self.old_action))
        delta = reward - self.function_approximation.summed_thetas(self.old_state, self.old_action)
        Qs = [self.function_approximation.summed_thetas(state, action) for action in self.actions]
        action, Q = self.policy.select_action(Qs)
        delta += self.gamma * Q
        self.function_approximation.set_thetas(self.alpha * delta * self.traces.traces)
        self.traces.decay()
        self.old_state = state
        self.old_action = action
        return action

    def reset(self, policy, traces, function_approximation, start_state, start_action):
        """Reset the policy, traces, function approximation, start action and/or start state"""
        self.policy = policy
        self.traces = traces
        self.function_approximation = function_approximation
        self.old_state = start_state
        self.old_action = start_action



# -*- coding: utf8 -*-

import os
import numpy as np
import tensorflow as tf

from yarll.agents.agent import Agent
from yarll.agents.env_runner import EnvRunner
from yarll.misc.utils import discount_rewards, FastSaver
from yarll.misc.reporter import Reporter
from yarll.misc.network_ops import create_accumulative_gradients_op, add_accumulative_gradients_op, reset_accumulative_gradients_op

class TaskPolicy(object):
    """Policy for a specific class."""
    def __init__(self, action, master):
        super(TaskPolicy, self).__init__()
        self.action = action
        self.master = master

    def choose_action(self, state):
        """Choose an action."""
        return self.master.session.run([self.action], feed_dict={self.master.states: [state]})[0]

    def new_trajectory(self):
        pass

class KnowledgeTransfer(Agent):
    """Learner for variations of a task."""
    def __init__(self, envs, monitor_path, **usercfg):
        super(KnowledgeTransfer, self).__init__(**usercfg)
        self.envs = envs
        self.n_tasks = len(envs)
        self.monitor_path = monitor_path
        self.nA = envs[0].action_space.n
        self.config.update(dict(
            timesteps_per_batch=10000,
            trajectories_per_batch=10,
            batch_update="timesteps",
            n_iter=100,
            switch_at_iter=None,
            gamma=0.99,  # Discount past rewards by a percentage
            decay=0.9,  # Decay of RMSProp optimizer
            epsilon=1e-9,  # Epsilon of RMSProp optimizer
            learning_rate=0.005,
            n_hidden_units=10,
            repeat_n_actions=1,
            n_sparse_units=10,
            feature_extraction=False
        ))
        self.config.update(usercfg)

        self.build_networks()
        self.task_runners = [EnvRunner(envs[i], TaskPolicy(action, self), self.config) for i, action in enumerate(self.action_tensors)]
        if self.config["save_model"]:
            for action_tensor in self.action_tensors:
                tf.add_to_collection("action", action_tensor)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()

    def build_networks(self):
        self.session = tf.Session()

        with tf.variable_scope("shared"):
            self.states = tf.placeholder(tf.float32, [None] + list(self.envs[0].observation_space.shape), name="states")
            self.action_taken = tf.placeholder(tf.float32, name="action_taken")
            self.advantage = tf.placeholder(tf.float32, name="advantage")

            L1 = None
            if self.config["feature_extraction"]:
                L1 = tf.contrib.layers.fully_connected(
                    inputs=self.states,
                    num_outputs=self.config["n_hidden_units"],
                    activation_fn=tf.tanh,
                    weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
                    biases_initializer=tf.zeros_initializer(),
                    scope="L1")
            else:
                L1 = self.states

            knowledge_base = tf.Variable(tf.truncated_normal([L1.get_shape()[-1].value, self.config["n_sparse_units"]], mean=0.0, stddev=0.02), name="knowledge_base")

            self.shared_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope="shared")

        # Every task has its own (sparse) representation
        sparse_representations = [
            tf.Variable(tf.truncated_normal([self.config["n_sparse_units"], self.nA], mean=0.0, stddev=0.02), name="sparse%d" % i)
            for i in range(self.n_tasks)
        ]

        self.probs_tensors = [tf.nn.softmax(tf.matmul(L1, tf.matmul(knowledge_base, s))) for s in sparse_representations]
        self.action_tensors = [tf.squeeze(tf.multinomial(tf.log(probs), 1)) for probs in self.probs_tensors]

        self.optimizer = tf.train.RMSPropOptimizer(
            learning_rate=self.config["learning_rate"],
            decay=self.config["decay"],
            epsilon=self.config["epsilon"]
        )
        net_vars = self.shared_vars + sparse_representations
        self.accum_grads = create_accumulative_gradients_op(net_vars, 0)

        self.loss = tf.placeholder("float", name="loss")
        summary_loss = tf.summary.scalar("Loss", self.loss)
        self.rewards = tf.placeholder("float", name="Rewards")
        summary_rewards = tf.summary.scalar("Reward", self.rewards)
        self.episode_lengths = tf.placeholder("float", name="Episode_lengths")
        summary_episode_lengths = tf.summary.scalar("Length", self.episode_lengths)
        self.summary_op = tf.summary.merge([summary_loss, summary_rewards, summary_episode_lengths])

        self.writers = []
        self.losses = []

        regularizer = tf.contrib.layers.l1_regularizer(.05)
        for i, probabilities in enumerate(self.probs_tensors):
            good_probabilities = tf.reduce_sum(tf.multiply(probabilities, tf.one_hot(tf.cast(self.action_taken, tf.int32), self.nA)),
                                               reduction_indices=[1])
            eligibility = tf.log(good_probabilities) * self.advantage
            loss = -tf.reduce_sum(eligibility) + regularizer(sparse_representations[i])
            self.losses.append(loss)
            writer = tf.summary.FileWriter(os.path.join(self.monitor_path, "task" + str(i)), self.session.graph)
            self.writers.append(writer)

        # An add op for every task & its loss
        self.add_accum_grads = []
        for i, loss in enumerate(self.losses):
            # Use all variables if the switch tasks experiment is disactivated or it's not the last task
            all_vars = self.config["switch_at_iter"] is None or i != len(self.losses) - 1
            self.add_accum_grads.append(add_accumulative_gradients_op(
                (self.shared_vars if all_vars else []) + [sparse_representations[i]],
                ([self.accum_grads[0]] if all_vars else []) + [self.accum_grads[i + 1]],
                loss,
                i
            ))

        self.apply_gradients = self.optimizer.apply_gradients(
            zip(self.accum_grads, net_vars))
        self.reset_accum_grads = reset_accumulative_gradients_op(net_vars, self.accum_grads, 0)

        self.init_op = tf.global_variables_initializer()

    def _initialize(self):
        self.session.run(self.init_op)

    def learn(self):
        """Run learning algorithm"""
        self._initialize()
        reporter = Reporter()
        config = self.config
        total_n_trajectories = np.zeros(len(self.envs))
        for iteration in range(config["n_iter"]):
            self.session.run([self.reset_accum_grads])
            for i, task_runner in enumerate(self.task_runners):
                if self.config["switch_at_iter"] is not None:
                    if iteration >= self.config["switch_at_iter"] and i != (len(self.task_runners) - 1):
                        continue
                    elif iteration < self.config["switch_at_iter"] and i == len(self.task_runners) - 1:
                        continue
                # Collect trajectories until we get timesteps_per_batch total timesteps
                trajectories = task_runner.get_trajectories()
                total_n_trajectories[i] += len(trajectories)
                all_state = np.concatenate([trajectory["state"] for trajectory in trajectories])
                # Compute discounted sums of rewards
                rets = [discount_rewards(trajectory["reward"], config["gamma"]) for trajectory in trajectories]
                max_len = max(len(ret) for ret in rets)
                padded_rets = [np.concatenate([ret, np.zeros(max_len - len(ret))]) for ret in rets]
                # Compute time-dependent baseline
                baseline = np.mean(padded_rets, axis=0)
                # Compute advantage function
                advs = [ret - baseline[:len(ret)] for ret in rets]
                all_action = np.concatenate([trajectory["action"] for trajectory in trajectories])
                all_adv = np.concatenate(advs)
                # Do policy gradient update step
                episode_rewards = np.array([trajectory["reward"].sum() for trajectory in trajectories])  # episode total rewards
                episode_lengths = np.array([len(trajectory["reward"]) for trajectory in trajectories])  # episode lengths
                results = self.session.run([self.losses[i], self.add_accum_grads[i], self.accum_grads], feed_dict={
                    self.states: all_state,
                    self.action_taken: all_action,
                    self.advantage: all_adv
                })
                summary = self.session.run([self.summary_op], feed_dict={
                    self.loss: results[0],
                    self.rewards: np.mean(episode_rewards),
                    self.episode_lengths: np.mean(episode_lengths)
                })

                self.writers[i].add_summary(summary[0], iteration)
                self.writers[i].flush()
                print("Task:", i)
                reporter.print_iteration_stats(iteration, episode_rewards, episode_lengths, total_n_trajectories[i])

            # Apply accumulated gradient after all the gradients of each task are summed
            self.session.run([self.apply_gradients])

        if self.config["save_model"]:
            if not os.path.exists(self.monitor_path):
                os.makedirs(self.monitor_path)
            self.saver.save(self.session, os.path.join(self.monitor_path, "model"))



# -*- coding: utf8 -*-

import numpy as np

class OrnsteinUhlenbeckActionNoise(object):
    """
    Ornstein Uhlenbeck process action noise.
    """

    def __init__(self, n_actions: int, sigma: float, theta: float = .15, dt: float = 1e-2, x0=None) -> None:
        super(OrnsteinUhlenbeckActionNoise, self).__init__()
        self.theta: float = theta
        self.mu: np.ndarray = np.zeros(n_actions)
        self.sigma: float = sigma
        self.dt: float = dt
        self.x0: float = x0
        self.x_prev = None
        self.reset()

    def __call__(self) -> np.ndarray:
        x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + \
            self.sigma * np.sqrt(self.dt) * \
            np.random.normal(size=self.mu.shape)
        self.x_prev = x
        return x

    def reset(self):
        self.x_prev = self.x0 if self.x0 is not None else np.zeros_like(self.mu)

    def __repr__(self) -> str:
        return 'OrnsteinUhlenbeckActionNoise(mu={}, sigma={})'.format(self.mu, self.sigma)



# -*- coding: utf8 -*-

import os
import numpy as np
import tensorflow as tf

from yarll.agents.agent import Agent
from yarll.memory.memory import Memory
from yarll.misc.network_ops import linear, normalized_columns_initializer

class SAC(Agent):
    def __init__(self, env, monitor_path: str, **usercfg) -> None:
        super(SAC, self).__init__(**usercfg)
        self.env = env
        self.monitor_path: str = monitor_path

        self.config.update(
            n_episodes=100000,
            n_timesteps=env.spec.tags.get("wrapper_config.TimeLimit.max_episode_steps"),
            actor_learning_rate=3e-4,
            softq_learning_rate=3e-4,
            value_learning_rate=3e-4,
            gamma=0.99,
            batch_size=128,
            tau=0.01,
            n_actor_layers=2,
            logprob_epsilon=1e-6, # For numerical stability when computing tf.log
            n_hidden_units=128,
            n_train_steps=4, # Number of parameter update steps per iteration
            replay_buffer_size=1e6,
            replay_start_size=128  # Required number of replay buffer entries to start training
        )
        self.config.update(usercfg)

        self.state_shape: list = list(env.observation_space.shape)
        self.n_actions: int = env.action_space.shape[0]

        # Placeholders
        self.states = tf.placeholder(tf.float32, [None] + self.state_shape, name="states")
        self.actions_taken = tf.placeholder(tf.float32, [self.config["batch_size"], self.n_actions], name="actions_taken")
        self.softq_target = tf.placeholder(tf.float32, [self.config["batch_size"], 1], name="softq_target")

        # Make networks
        # action_output are the squashed actions and action_original those straight from the normal distribution
        self.action_output, self.action_original, self.action_logprob, self.actor_vars = self.build_actor_network()
        self.softq_output, self.softq_vars = self.build_softq_network(self.actions_taken)
        self.new_softq_output, _ = self.build_softq_network(self.action_output, reuse_vars=True)
        self.value_output, self.value_vars = self.build_value_network()
        self.target_value_output, self.value_target_update = self.build_target_value_network(self.value_vars)

        # Make losses
        self.softq_loss = tf.reduce_mean(tf.square(self.softq_output - self.softq_target), name="softq_loss")
        value_target = tf.stop_gradient(self.new_softq_output - self.action_logprob)
        self.value_loss = tf.reduce_mean(tf.square(self.value_output - value_target), name="value_loss")
        advantage = tf.stop_gradient(self.action_logprob - self.new_softq_output + self.value_output)
        self.actor_loss = tf.reduce_mean(self.action_logprob * advantage, name="actor_loss")

        # Make train ops
        self.softq_train_op = tf.train.AdamOptimizer(
            learning_rate=self.config["softq_learning_rate"],
            name="softq_optimizer").minimize(self.softq_loss)
        self.value_train_op = tf.train.AdamOptimizer(
            learning_rate=self.config["value_learning_rate"],
            name="value_optimizer").minimize(self.value_loss)

        self.actor_train_op = tf.train.AdamOptimizer(
            learning_rate=self.config["actor_learning_rate"],
            name="actor_optimizer").minimize(self.actor_loss)

        summaries = []
        for v in self.actor_vars + self.softq_vars + self.value_vars:
            summaries.append(tf.summary.histogram(v.name, v))

        self.model_summary_op = tf.summary.merge(summaries)

        self.init_op = tf.global_variables_initializer()

        self.replay_buffer = Memory(int(self.config["replay_buffer_size"]))

        self.n_updates = 0

        self.summary_writer = tf.summary.FileWriter(os.path.join(
            self.monitor_path, "summaries"), tf.get_default_graph())

    def build_actor_network(self):
        w_bound = 3e-3
        x = self.states
        with tf.variable_scope("actor"):
            x = tf.nn.relu(linear(x, self.config["n_hidden_units"], "L1", normalized_columns_initializer()))
            x = tf.nn.relu(linear(x, self.config["n_hidden_units"], "L2", normalized_columns_initializer()))

            mean = linear(x,
                          self.n_actions,
                          "mean",
                          tf.random_uniform_initializer(-w_bound, w_bound),
                          tf.random_uniform_initializer(-w_bound, w_bound))
            log_std = linear(x,
                             self.n_actions,
                             "log_std",
                             tf.random_uniform_initializer(-w_bound, w_bound),
                             tf.random_uniform_initializer(-w_bound, w_bound))
            log_std_clipped = tf.clip_by_value(log_std, -20, 2, name="log_std_clipped") # In autor's code but not in paper

            normal_dist = tf.distributions.Normal(mean, tf.exp(log_std_clipped), name="actions_normal_distr")
            actions = tf.stop_gradient(normal_dist.sample(name="actions"))
            squashed_actions = tf.tanh(actions, name="squashed_actions")  # Squash output between [-1, 1]

            logprob = normal_dist.log_prob(actions) - \
            tf.log(1.0 - tf.pow(squashed_actions, 2) + self.config["logprob_epsilon"])
            logprob = tf.reduce_sum(logprob, axis=-1, keepdims=True)

            actor_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

        return squashed_actions, actions, logprob, actor_vars

    def build_softq_network(self, actions, reuse_vars=False):
        x = tf.concat([self.states, actions], 1)
        with tf.variable_scope("softq") as scope:
            if reuse_vars:
                scope.reuse_variables()
            x = tf.nn.relu(linear(x, self.config["n_hidden_units"], "L1", normalized_columns_initializer()))
            x = tf.nn.relu(linear(x, self.config["n_hidden_units"], "L2", normalized_columns_initializer()))
            x = linear(x,
                       1,
                       "output",
                       tf.random_uniform_initializer(-3e-3, 3e-3),
                       tf.random_uniform_initializer(-3e-3, 3e-3))

            softq_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

        return x, softq_vars

    def build_value_network(self):
        x = self.states
        with tf.variable_scope("value"):
            x = tf.nn.relu(linear(x, self.config["n_hidden_units"], "L1", normalized_columns_initializer()))
            x = tf.nn.relu(linear(x, self.config["n_hidden_units"], "L2", normalized_columns_initializer()))
            x = linear(x,
                       1,
                       "output",
                       tf.random_uniform_initializer(-3e-3, 3e-3),
                       tf.random_uniform_initializer(-3e-3, 3e-3))

            value_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

        return x, value_vars

    def build_target_value_network(self, value_vars: list):
        ema = tf.train.ExponentialMovingAverage(decay=1 - self.config["tau"])
        target_update = ema.apply(value_vars)
        target_net = [ema.average(v) for v in value_vars]

        x = self.states
        x = tf.nn.relu(tf.nn.xw_plus_b(x, target_net[0], target_net[1]))
        x = tf.nn.relu(tf.nn.xw_plus_b(x, target_net[2], target_net[3]))
        value = tf.nn.xw_plus_b(x, target_net[4], target_net[5], name="target_value")

        return value, target_update

    def value(self, states: np.ndarray):
        return tf.get_default_session().run(self.value_output, feed_dict={
            self.states: states
        })

    def target_value(self, states: np.ndarray):
        return tf.get_default_session().run(self.target_value_output, feed_dict={
            self.states: states
        })

    def softq_value(self, states: np.ndarray, actions: np.ndarray):
        return tf.get_default_session().run(self.softq_output, feed_dict={
            self.states: states,
            self.actions_taken: actions
            })

    def actions(self, states: np.ndarray) -> np.ndarray:
        """Get the actions for a batch of states."""
        return tf.get_default_session().run(self.action_output, feed_dict={
            self.states: states
        })

    def action(self, state: np.ndarray) -> np.ndarray:
        """Get the action for a single state."""
        return self.actions([state])[0]

    def train(self, model_summary=True):
        sample = self.replay_buffer.get_batch(self.config["batch_size"])

        # for n_actions == 1
        action_batch = np.resize(sample["actions"], [self.config["batch_size"], self.n_actions])

        # Calculate critic targets
        next_value_batch = self.target_value(sample["states1"])
        softq_targets = sample["rewards"] + (1 - sample["terminals1"]) * \
            self.config["gamma"] * next_value_batch.squeeze()
        softq_targets = np.resize(softq_targets, [self.config["batch_size"], 1]).astype(np.float32)

        fetches = [self.action_logprob, self.actor_train_op, self.value_train_op]
        if model_summary:
            fetches += [self.actor_loss, self.value_loss]
        results = tf.get_default_session().run(
            fetches,
            feed_dict={self.states: sample["states0"]}
        )

        logprob = results[0]

        if model_summary:
            actor_loss = results[-2]
            value_loss = results[-1]

        fetches = [self.softq_train_op]
        if model_summary:
            fetches = [self.softq_output, self.softq_loss] + fetches
        results = tf.get_default_session().run(fetches, feed_dict={
            self.softq_target: softq_targets,
            self.states: sample["states0"],
            self.actions_taken: action_batch
        })

        if model_summary:
            summary = tf.Summary()
            summary.value.add(tag="model/predicted_softq_mean", simple_value=np.mean(results[0]))
            summary.value.add(tag="model/predicted_softq_std", simple_value=np.std(results[0]))
            summary.value.add(tag="model/softq_loss", simple_value=float(results[1]))
            summary.value.add(tag="model/actor_loss", simple_value=float(actor_loss))
            summary.value.add(tag="model/value_loss", simple_value=float(value_loss))
            summary.value.add(tag="model/action_logprob_mean", simple_value=np.mean(logprob))
            self.summary_writer.add_summary(summary, self.n_updates)

        # Update the target networks
        fetches = [self.value_target_update]
        if model_summary:
            fetches = [self.model_summary_op] + fetches
        res = tf.get_default_session().run(fetches)
        if model_summary:
            self.summary_writer.add_summary(res[0], self.n_updates)
        self.n_updates += 1

    def learn(self):
        action_low = self.env.action_space.low
        action_high = self.env.action_space.high
        with tf.Session() as sess, sess.as_default():
            sess.run(self.init_op)
            for episode in range(self.config["n_episodes"]):
                state = self.env.reset()
                episode_reward = 0
                episode_length = 0
                for _ in range(self.config["n_timesteps"]):
                    action = self.action(state)
                    to_execute = action_low + (action + 1.0) * 0.5 * (action_high - action_low)
                    new_state, reward, done, _ = self.env.step(to_execute)
                    episode_length += 1
                    episode_reward += reward
                    self.replay_buffer.add(state, action, reward, new_state, done)
                    if self.replay_buffer.n_entries > self.config["replay_start_size"]:
                        for _ in range(self.config["n_train_steps"]):
                            self.train(model_summary=True)
                    state = new_state
                    if done:
                        summary = tf.Summary()
                        summary.value.add(tag="global/Episode_length",
                                          simple_value=float(episode_length))
                        summary.value.add(tag="global/Reward",
                                          simple_value=float(episode_reward))
                        self.summary_writer.add_summary(summary, episode)
                        self.summary_writer.flush()
                        break



# -*- coding: utf8 -*-

import os
import tensorflow as tf
import numpy as np
from gym import wrappers

from yarll.agents.agent import Agent
from yarll.agents.actorcritic.actor_critic import ActorCriticNetwork, ActorCriticNetworkDiscrete,\
ActorCriticNetworkDiscreteCNN, ActorCriticNetworkContinuous
from yarll.agents.env_runner import EnvRunner
from yarll.misc.utils import FastSaver

def ppo_loss(old_log, new_log, epsilon, advantage):
    ratio = tf.exp(new_log - old_log)
    ratio_clipped = tf.clip_by_value(ratio, 1.0 - epsilon, 1.0 + epsilon)
    return tf.minimum(ratio * advantage, ratio_clipped * advantage)

class PPO(Agent):
    """Proximal Policy Optimization agent."""
    RNN = False

    def __init__(self, env, monitor_path: str, monitor: bool = False, video: bool = False, **usercfg) -> None:
        super(PPO, self).__init__(**usercfg)
        self.monitor_path: str = monitor_path
        self.env = env
        if monitor:
            self.env = wrappers.Monitor(
                self.env,
                monitor_path,
                force=True,
                video_callable=(None if video else False))

        self.config.update(dict(
            n_hidden_units=20,
            n_hidden_layers=2,
            gamma=0.99,
            gae_lambda=0.95,
            learning_rate=0.001,
            n_epochs=10,
            n_iter=10000,
            batch_size=64,  # Timesteps per training batch
            n_local_steps=256,
            normalize_states=False,
            gradient_clip_value=None,
            adam_epsilon=1e-5,
            vf_coef=0.5,
            entropy_coef=0.01,
            cso_epsilon=0.2,  # Clipped surrogate objective epsilon
            save_model=False
        ))
        self.config.update(usercfg)

        with tf.variable_scope("old_network"):
            self.old_network = self.build_networks()
            self.old_network_vars = tf.get_collection(
                tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

        with tf.variable_scope("new_network"):
            self.new_network = self.build_networks()
            if self.RNN:
                self.initial_features = self.new_network.state_init
            else:
                self.initial_features = None
            self.new_network_vars = tf.get_collection(
                tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)
        self.action = self.new_network.action
        self.value = self.new_network.value
        self.states = self.new_network.states
        self.actions_taken = self.new_network.actions_taken
        self.advantage = tf.placeholder(tf.float32, [None], name="advantage")
        self.ret = tf.placeholder(tf.float32, [None], name="return")

        self.set_old_to_new = tf.group(
            *[v1.assign(v2) for v1, v2 in zip(self.old_network_vars, self.new_network_vars)])

        self.actor_loss = -tf.reduce_mean(self.make_actor_loss(self.old_network, self.new_network, self.advantage))
        self.critic_loss = tf.reduce_mean(tf.square(self.value - self.ret))
        self.mean_entropy = tf.reduce_mean(self.new_network.entropy)
        self.loss = self.actor_loss + self.config["vf_coef"] * self.critic_loss + \
            self.config["entropy_coef"] * self.mean_entropy

        grads = tf.gradients(self.loss, self.new_network_vars)

        self._global_step = tf.get_variable(
            "global_step",
            [],
            tf.int32,
            initializer=tf.constant_initializer(0, dtype=tf.int32),
            trainable=False)

        self.n_steps = tf.shape(self.states)[0]
        self.session = tf.Session()
        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()

        summary_actor_loss = tf.summary.scalar("model/Actor_loss", self.actor_loss)
        summary_critic_loss = tf.summary.scalar("model/Critic_loss", self.critic_loss)
        summary_loss = tf.summary.scalar("model/Loss", self.loss)

        adv_mean, adv_std = tf.nn.moments(self.advantage, axes=[0])
        summary_adv_mean = tf.summary.scalar("model/advantage/mean", adv_mean)
        summary_adv_std = tf.summary.scalar("model/advantage/std", adv_std)

        # TODO: get from ppo_loss function
        # ratio_mean, ratio_std = tf.nn.moments(ratio, axes=[0])
        # summary_ratio_mean = tf.summary.scalar("model/ratio/mean", ratio_mean)
        # summary_ratio_std = tf.summary.scalar("model/ratio/std", ratio_std)

        summary_new_log_prob_mean = tf.summary.scalar(
            "model/new_log_prob/mean", tf.reduce_mean(self.new_network.action_log_prob))
        summary_old_log_prob_mean = tf.summary.scalar(
            "model/old_log_prob/mean", tf.reduce_mean(self.old_network.action_log_prob))

        summary_ret = tf.summary.scalar("model/return/mean", tf.reduce_mean(self.ret))
        summary_entropy = tf.summary.scalar("model/entropy", -self.mean_entropy)
        summary_grad_norm = tf.summary.scalar("model/grad_global_norm", tf.global_norm(grads))
        summary_var_norm = tf.summary.scalar(
            "model/var_global_norm", tf.global_norm(self.new_network_vars))
        summaries = []
        for v in tf.trainable_variables():
            if "new_network" in v.name:
                summaries.append(tf.summary.histogram(v.name, v))
        summaries += [summary_actor_loss, summary_critic_loss,
                      summary_loss,
                      summary_adv_mean, summary_adv_std,
                      # summary_ratio_mean, summary_ratio_std,
                      summary_new_log_prob_mean, summary_old_log_prob_mean,
                      summary_ret, summary_entropy, summary_grad_norm, summary_var_norm]
        self.model_summary_op = tf.summary.merge(summaries)
        self.writer = tf.summary.FileWriter(os.path.join(
            self.monitor_path, "summaries"), self.session.graph)
        self.env_runner = EnvRunner(self.env,
                                    self,
                                    usercfg,
                                    normalize_states=self.config["normalize_states"],
                                    summary_writer=self.writer)

        # grads before clipping were passed to the summary, now clip and apply them
        if self.config["gradient_clip_value"] is not None:
            grads, _ = tf.clip_by_global_norm(
                grads, self.config["gradient_clip_value"])
        self.optimizer = tf.train.AdamOptimizer(
            learning_rate=self.config["learning_rate"],
            epsilon=self.config["adam_epsilon"],
            name="optim")
        apply_grads = self.optimizer.apply_gradients(
            zip(grads, self.new_network_vars))

        inc_step = self._global_step.assign_add(self.n_steps)
        self.train_op = tf.group(apply_grads, inc_step)

        self.init_op = tf.global_variables_initializer()
        return

    def _initialize(self):
        self.session.run(self.init_op)

    def make_actor_loss(self, old_network, new_network, advantage):
        return ppo_loss(old_network.action_log_prob, new_network.action_log_prob, self.config["cso_epsilon"], advantage)

    def build_networks(self):
        raise NotImplementedError

    @property
    def global_step(self):
        return self._global_step.eval(session=self.session)

    def get_critic_value(self, state, *rest):
        return self.session.run([self.value], feed_dict={self.states: state})[0].flatten()

    def choose_action(self, state, *rest):
        action, value = self.session.run(
            [self.action, self.value], feed_dict={self.states: [state]})
        return {"action": action, "value": value[0]}

    def get_env_action(self, action):
        return np.argmax(action)

    def get_processed_trajectories(self):
        experiences = self.env_runner.get_steps(
            int(self.config["n_local_steps"]), stop_at_trajectory_end=False)
        T = experiences.steps
        v = 0 if experiences.terminals[-1] else self.get_critic_value(
            np.asarray(experiences.states)[None, -1], experiences.features[-1])
        vpred = np.asarray(experiences.values + [v])
        gamma = self.config["gamma"]
        lambda_ = self.config["gae_lambda"]
        gaelam = advantages = np.empty(T, 'float32')
        last_gaelam = 0
        for t in reversed(range(T)):
            nonterminal = 1 - experiences.terminals[t]
            delta = experiences.rewards[t] + gamma * vpred[t + 1] * nonterminal - vpred[t]
            gaelam[t] = last_gaelam = delta + gamma * lambda_ * nonterminal * last_gaelam
        rs = advantages + experiences.values
        return experiences.states, experiences.actions, advantages, rs, experiences.features

    def learn(self):
        """Run learning algorithm"""
        self._initialize()
        config = self.config
        n_updates = 0
        for _ in range(int(config["n_iter"])):
            # Collect trajectories until we get timesteps_per_batch total timesteps
            states, actions, advs, rs, _ = self.get_processed_trajectories()
            advs = np.array(advs)
            advs = (advs - advs.mean()) / advs.std()
            self.session.run(self.set_old_to_new)

            indices = np.arange(len(states))
            for _ in range(int(self.config["n_epochs"])):
                np.random.shuffle(indices)

                batch_size = int(self.config["batch_size"])
                for j in range(0, len(states), batch_size):
                    batch_indices = indices[j:(j + batch_size)]
                    batch_states = np.array(states)[batch_indices]
                    batch_actions = np.array(actions)[batch_indices]
                    batch_advs = np.array(advs)[batch_indices]
                    batch_rs = np.array(rs)[batch_indices]
                    losses = [self.actor_loss, self.critic_loss, self.loss]
                    fetches = losses + [self.model_summary_op, self.train_op]
                    feed_dict = {
                        self.states: batch_states,
                        self.old_network.states: batch_states,
                        self.actions_taken: batch_actions,
                        self.old_network.actions_taken: batch_actions,
                        self.advantage: batch_advs,
                        self.ret: batch_rs
                    }
                    results = self.session.run(fetches, feed_dict)
                    self.writer.add_summary(results[len(losses)], n_updates)
                    n_updates += 1
                self.writer.flush()

            if self.config["save_model"]:
                self.saver.save(self.session, os.path.join(self.monitor_path, "model"))


class PPODiscrete(PPO):
    def build_networks(self) -> ActorCriticNetwork:
        return ActorCriticNetworkDiscrete(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))


class PPODiscreteCNN(PPODiscrete):
    def build_networks(self) -> ActorCriticNetwork:
        return ActorCriticNetworkDiscreteCNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]))


class PPOContinuous(PPO):
    def build_networks(self) -> ActorCriticNetwork:
        return ActorCriticNetworkContinuous(
            list(self.env.observation_space.shape),
            self.env.action_space,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))

    def get_env_action(self, action):
        return action



# -*- coding: utf8 -*-

import numpy as np

class EligibilityTraces(object):
    """Eligibility traces"""
    def __init__(self, features_shape, gamma: float, Lambda: float) -> None:
        super(EligibilityTraces, self).__init__()
        self.features_shape = features_shape
        self.traces = np.zeros(self.features_shape)
        self.gamma: float = gamma
        self.Lambda: float = Lambda

    def replacing_traces(self, present_features):
        self.traces = present_features  # replacing traces: set them to 1

    def decay(self):
        """Reduce the traces by taking a fraction of them, determined by gamma and lambda"""
        self.traces *= self.gamma * self.Lambda



# -*- coding: utf8 -*-

from gym import wrappers

from yarll.policies.e_greedy import EGreedy
from yarll.agents.sarsa import Sarsa
from yarll.traces.eligibility_traces import EligibilityTraces
from yarll.functionapproximation.tile_coding import TileCoding

# def draw_3d(tile_starts):
#     states = []
#     for i in range(n_x_tiles):
#         for j in range(n_y_tiles):
#             states.append([i, j])
#     states = np.array(states)

class SarsaFA(object):
    """Learner using Sarsa and function approximation"""
    def __init__(self, env, monitor_path: str, video: bool = True, **usercfg) -> None:
        super(SarsaFA, self).__init__()
        self.env = env
        self.env = wrappers.Monitor(self.env, monitor_path, force=True, video_callable=(None if video else False))
        m = usercfg.get("m", 10)  # Number of tilings
        self.config = dict(
            m=m,
            n_x_tiles=9,
            n_y_tiles=9,
            Lambda=0.9,
            epsilon=0,  # fully greedy in this case
            alpha=(0.05 * (0.5 / m)),
            gamma=1,
            n_iter=1000,
            steps_per_episode=env.spec.tags.get("wrapper_config.TimeLimit.max_episode_steps")
        )
        self.config.update(usercfg)
        O = env.observation_space
        self.x_low, self.y_low = O.low
        self.x_high, self.y_high = O.high

        self.nA = env.action_space.n
        self.policy = EGreedy(self.config["epsilon"])
        self.function_approximation = TileCoding(self.x_low, self.x_high,
                                                 self.y_low, self.y_high,
                                                 m,
                                                 int(self.config["n_x_tiles"]), int(self.config["n_y_tiles"]),
                                                 self.nA)

    def learn(self):
        for i in range(int(self.config["n_iter"])):
            traces = EligibilityTraces(self.function_approximation.features_shape,
                                       self.config["gamma"],
                                       self.config["Lambda"])
            state, action = self.env.reset(), 0
            sarsa = Sarsa(self.config["gamma"],
                          self.config["alpha"],
                          self.policy,
                          traces,
                          self.function_approximation,
                          range(self.nA), state, action)
            done = False
            iteration = 0
            while not done:
                iteration += 1
                state, reward, done, _ = self.env.step(action)
                if done and iteration < self.config["steps_per_episode"]:
                    print("Episode {}: Less than {} steps were needed: {}".format(i,
                                                                                  self.config["steps_per_episode"],
                                                                                  iteration))
                action = sarsa.step(state, reward)



# -*- coding: utf8 -*-

import os
import logging
import tensorflow as tf
import numpy as np

from gym import wrappers

from yarll.agents.agent import Agent
from yarll.agents.actorcritic.actor_critic import ActorCriticNetworkDiscrete,\
ActorCriticNetworkDiscreteCNN, ActorCriticNetworkDiscreteCNNRNN, actor_critic_discrete_loss,\
ActorCriticNetworkContinuous, actor_critic_continuous_loss
from yarll.agents.env_runner import EnvRunner
from yarll.misc.utils import discount_rewards, FastSaver

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Don't use the scientific notation to print results
np.set_printoptions(suppress=True)


class A2C(Agent):
    """Advantage Actor Critic"""

    def __init__(self, env, monitor_path: str, video: bool = True, **usercfg) -> None:
        super(A2C, self).__init__(**usercfg)
        self.monitor_path = monitor_path

        self.env = wrappers.Monitor(
            env,
            monitor_path,
            force=True,
            video_callable=(None if video else False))

        self.config.update(dict(
            n_iter=100,
            gamma=0.99,
            learning_rate=0.001,
            n_hidden_units=20,
            n_hidden_layers=1,
            gradient_clip_value=0.5,
            n_local_steps=20,
            vf_coef=0.5,
            entropy_coef=0.01,
            loss_reducer="mean",
            save_model=False
        ))
        self.config.update(usercfg)
        # Only used (and overwritten) by agents that use an RNN
        self.initial_features = None
        self.ac_net = None  # Overwritten by build_networks
        self.build_networks()

        self.action = self.ac_net.action
        self.states = self.ac_net.states
        self.actions_taken = self.ac_net.actions_taken
        self.advantage = tf.placeholder(tf.float32, [None], name="advantage")
        self.ret = tf.placeholder(tf.float32, [None], name="return")

        self.actor_loss, self.critic_loss, self.loss = self.make_loss()

        self.vars = tf.get_collection(
            tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

        self._global_step = tf.get_variable(
            "global_step",
            [],
            tf.int32,
            initializer=tf.constant_initializer(0, dtype=tf.int32),
            trainable=False)

        self.optimizer = tf.train.AdamOptimizer(
            self.config["learning_rate"], name="optim")
        grads = tf.gradients(self.loss, self.vars)
        grads, _ = tf.clip_by_global_norm(
            grads, self.config["gradient_clip_value"])

        # Apply gradients to the weights of the master network
        apply_grads = self.optimizer.apply_gradients(zip(grads, self.vars))

        self.n_steps = tf.shape(self.states)[0]
        inc_step = self._global_step.assign_add(self.n_steps)
        self.train_op = tf.group(apply_grads, inc_step)

        self.init_op = tf.global_variables_initializer()
        # Launch the graph.
        self.session = tf.Session()
        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()
        n_steps = tf.to_float(self.n_steps)
        actor_loss_summary = tf.summary.scalar("model/actor_loss", tf.squeeze(self.actor_loss / n_steps))
        critic_loss_summary = tf.summary.scalar("model/critic_loss", tf.squeeze(self.critic_loss / n_steps))
        loss_summary = tf.summary.scalar("model/loss", tf.squeeze(self.loss / n_steps))
        self.loss_summary_op = tf.summary.merge(
            [actor_loss_summary, critic_loss_summary, loss_summary])
        self.writer = tf.summary.FileWriter(os.path.join(
            self.monitor_path, "summaries"), self.session.graph)
        self.env_runner = EnvRunner(self.env, self, usercfg, summary_writer=self.writer)
        return

    def _initialize(self):
        self.session.run(self.init_op)

    def build_networks(self):
        return NotImplementedError("Abstract method")

    def make_loss(self):
        return NotImplementedError("Abstract method")

    @property
    def global_step(self):
        return self._global_step.eval(session=self.session)

    def get_critic_value(self, state, features):
        return self.session.run([self.ac_net.value], feed_dict={self.states: state})[0].flatten()

    def choose_action(self, state, features) -> dict:
        action, value = self.session.run(
            [self.ac_net.action, self.ac_net.value], feed_dict={self.states: [state]})
        return {"action": action, "value": value[0]}

    def get_env_action(self, action) -> int:
        return np.argmax(action)

    def learn(self):
        """Run learning algorithm"""
        self._initialize()
        config = self.config
        for _ in range(int(config["n_iter"])):
            # Collect trajectories until we get timesteps_per_batch total timesteps
            trajectory = self.env_runner.get_steps(int(self.config["n_local_steps"]))
            v = 0 if trajectory.terminals[-1] else self.get_critic_value(
                np.asarray(trajectory.states)[None, -1], trajectory.features[-1])
            rewards_plus_v = np.asarray(trajectory.rewards + [v])
            vpred_t = np.asarray(trajectory.values + [v])
            delta_t = trajectory.rewards + \
                self.config["gamma"] * vpred_t[1:] - vpred_t[:-1]
            batch_r = discount_rewards(
                rewards_plus_v, self.config["gamma"])[:-1]
            batch_adv = discount_rewards(delta_t, self.config["gamma"])
            fetches = [self.loss_summary_op, self.train_op, self._global_step]
            states = np.asarray(trajectory.states)
            feed_dict = {
                self.states: states,
                self.actions_taken: np.asarray(trajectory.actions),
                self.advantage: batch_adv,
                self.ret: np.asarray(batch_r)
            }
            feature = trajectory.features[0]
            if feature != [] and feature is not None:
                feed_dict[self.ac_net.rnn_state_in] = feature
            summary, _, global_step = self.session.run(fetches, feed_dict)
            self.writer.add_summary(summary, global_step)
            self.writer.flush()

        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver.save(self.session, os.path.join(
                self.monitor_path, "model"))


class A2CDiscrete(A2C):
    def __init__(self, *args, **kwargs):
        super(A2CDiscrete, self).__init__(*args, **kwargs)

    def build_networks(self):
        self.ac_net = ActorCriticNetworkDiscrete(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))

    def make_loss(self):
        return actor_critic_discrete_loss(
            self.ac_net.logits,
            self.ac_net.probs,
            self.ac_net.value,
            self.ac_net.actions_taken,
            self.advantage,
            self.ret,
            self.config["vf_coef"],
            self.config["entropy_coef"],
            self.config["loss_reducer"]
        )


class A2CDiscreteCNN(A2CDiscrete):
    def build_networks(self):
        self.ac_net = ActorCriticNetworkDiscreteCNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]))


class A2CDiscreteCNNRNN(A2CDiscrete):
    def build_networks(self):
        self.ac_net = ActorCriticNetworkDiscreteCNNRNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]))
        self.initial_features = self.ac_net.state_init

    def choose_action(self, state, features) -> dict:
        """Choose an action."""
        feed_dict = {
            self.ac_net.states: [state],
            self.ac_net.rnn_state_in: features
        }

        action, rnn_state, value = self.session.run(
            [self.ac_net.action, self.ac_net.rnn_state_out, self.ac_net.value],
            feed_dict=feed_dict)
        return {"action": action, "value": value[0], "features": rnn_state}

    def get_critic_value(self, states, features):
        feed_dict = {
            self.ac_net.states: states,
            self.ac_net.rnn_state_in: features
        }
        return self.session.run(self.ac_net.value, feed_dict=feed_dict)[0]


class A2CContinuous(A2C):
    def __init__(self, *args, **kwargs):
        super(A2CContinuous, self).__init__(*args, **kwargs)

    def build_networks(self):
        self.ac_net = ActorCriticNetworkContinuous(
            list(self.env.observation_space.shape),
            self.env.action_space,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))

    def make_loss(self):
        return actor_critic_continuous_loss(
            self.ac_net.action_log_prob,
            self.ac_net.entropy,
            self.ac_net.value,
            self.advantage,
            self.ret,
            self.config["vf_coef"],
            self.config["entropy_coef"],
            self.config["loss_reducer"]
        )

    def get_env_action(self, action):
        return action



# -*- coding: utf8 -*-

import os
import logging
import signal
from threading import Thread
import numpy as np
import tensorflow as tf

from yarll.agents.agent import Agent
from yarll.agents.env_runner import EnvRunner
from yarll.misc.utils import discount_rewards, FastSaver
from yarll.misc.reporter import Reporter
from yarll.agents.knowledgetransfer import TaskPolicy

class AKTThread(Thread):
    """Asynchronous knowledge transfer learner thread. Used to learn using one specific variation of a task."""
    def __init__(self, master, env, task_id, n_iter, start_at_iter=0):
        super(AKTThread, self).__init__()
        self.master = master
        self.config = self.master.config
        self.task_id = task_id
        self.nA = env.action_space.n
        self.n_iter = n_iter
        self.start_at_iter = start_at_iter
        self.add_accum_grad = None  # To be filled in later

        self.build_networks()
        self.states = self.master.states
        self.session = self.master.session
        self.task_runner = EnvRunner(env, TaskPolicy(self.action, self), self.master.config)

        # Write the summary of each task in a different directory
        self.writer = tf.summary.FileWriter(os.path.join(self.master.monitor_path, "task" + str(self.task_id)), self.master.session.graph)

        self.optimizer = tf.train.RMSPropOptimizer(learning_rate=self.config["learning_rate"], decay=self.config["decay"], epsilon=self.config["epsilon"])

    def build_networks(self):
        with tf.variable_scope("task{}".format(self.task_id)):
            self.sparse_representation = tf.Variable(tf.truncated_normal([self.master.config["n_sparse_units"], self.nA], mean=0.0, stddev=0.02))
            self.probs = tf.nn.softmax(tf.matmul(self.master.L1, tf.matmul(self.master.knowledge_base, self.sparse_representation)))

            self.action = tf.squeeze(tf.multinomial(tf.log(self.probs), 1), name="action")

            good_probabilities = tf.reduce_sum(tf.multiply(self.probs, tf.one_hot(tf.cast(self.master.action_taken, tf.int32), self.nA)), reduction_indices=[1])
            eligibility = tf.log(good_probabilities + 1e-10) * self.master.advantage
            self.loss = -tf.reduce_sum(eligibility)

    def run(self):
        """Run the appropriate learning algorithm."""
        if self.master.learning_method == "REINFORCE":
            self.learn_REINFORCE()
        else:
            self.learn_Karpathy()

    def learn_REINFORCE(self):
        """Learn using updates like in the REINFORCE algorithm."""
        reporter = Reporter()
        total_n_trajectories = 0
        iteration = self.start_at_iter
        while iteration < self.n_iter and not self.master.stop_requested:
            iteration += 1
            # Collect trajectories until we get timesteps_per_batch total timesteps
            trajectories = self.task_runner.get_trajectories()
            total_n_trajectories += len(trajectories)
            all_state = np.concatenate([trajectory["state"] for trajectory in trajectories])
            # Compute discounted sums of rewards
            rets = [discount_rewards(trajectory["reward"], self.config["gamma"]) for trajectory in trajectories]
            max_len = max(len(ret) for ret in rets)
            padded_rets = [np.concatenate([ret, np.zeros(max_len - len(ret))]) for ret in rets]
            # Compute time-dependent baseline
            baseline = np.mean(padded_rets, axis=0)
            # Compute advantage function
            advs = [ret - baseline[:len(ret)] for ret in rets]
            all_action = np.concatenate([trajectory["action"] for trajectory in trajectories])
            all_adv = np.concatenate(advs)
            # Do policy gradient update step
            episode_rewards = np.array([trajectory["reward"].sum() for trajectory in trajectories])  # episode total rewards
            episode_lengths = np.array([len(trajectory["reward"]) for trajectory in trajectories])  # episode lengths
            results = self.master.session.run([self.loss, self.apply_grad], feed_dict={
                self.master.states: all_state,
                self.master.action_taken: all_action,
                self.master.advantage: all_adv
            })
            print("Task:", self.task_id)
            reporter.print_iteration_stats(iteration, episode_rewards, episode_lengths, total_n_trajectories)
            summary = self.master.session.run([self.master.summary_op], feed_dict={
                self.master.loss: results[0],
                self.master.reward: np.mean(episode_rewards),
                self.master.episode_length: np.mean(episode_lengths)
            })
            self.writer.add_summary(summary[0], iteration)
            self.writer.flush()

    def learn_Karpathy(self):
        """Learn using updates like in the Karpathy algorithm."""
        iteration = self.start_at_iter
        while iteration < self.n_iter and not self.master.stop_requested:  # Keep executing episodes until the master requests a stop (e.g. using SIGINT)
            iteration += 1
            trajectory = self.task_runner.get_trajectory()
            reward = sum(trajectory["reward"])
            action_taken = trajectory["action"]

            discounted_episode_rewards = discount_rewards(trajectory["reward"], self.config["gamma"])
            # standardize
            discounted_episode_rewards -= np.mean(discounted_episode_rewards)
            std = np.std(discounted_episode_rewards)
            std = std if std > 0 else 1
            discounted_episode_rewards /= std
            feedback = discounted_episode_rewards

            results = self.master.session.run([self.loss, self.apply_grad], feed_dict={
                self.master.states: trajectory["state"],
                self.master.action_taken: action_taken,
                self.master.advantage: feedback
            })
            results = self.master.session.run([self.master.summary_op], feed_dict={
                self.master.loss: results[0],
                self.master.reward: reward,
                self.master.episode_length: trajectory["steps"]
            })
            self.writer.add_summary(results[0], iteration)
            self.writer.flush()


class AsyncKnowledgeTransfer(Agent):
    """Asynchronous learner for variations of a task."""
    def __init__(self, envs, monitor_path, learning_method="REINFORCE", **usercfg):
        super(AsyncKnowledgeTransfer, self).__init__(**usercfg)
        self.envs = envs
        self.learning_method = learning_method
        self.monitor_path = monitor_path
        self.config.update(dict(
            timesteps_per_batch=10000,
            trajectories_per_batch=10,
            batch_update="timesteps",
            n_iter=200,
            switch_at_iter=None,  # None to deactivate, otherwhise an iteration at which to switch
            gamma=0.99,  # Discount past rewards by a percentage
            decay=0.9,  # Decay of RMSProp optimizer
            epsilon=1e-9,  # Epsilon of RMSProp optimizer
            learning_rate=0.005,
            n_hidden_units=10,
            repeat_n_actions=1,
            n_task_variations=3,
            n_sparse_units=10,
            feature_extraction=False
        ))
        self.config.update(usercfg)

        self.stop_requested = False

        self.session = tf.Session(config=tf.ConfigProto(
            log_device_placement=False,
            allow_soft_placement=True))

        self.global_step = tf.get_variable("global_step", [], tf.int32, initializer=tf.constant_initializer(0, dtype=tf.int32), trainable=False)
        self.build_networks()

        self.loss = tf.placeholder("float", name="loss")
        summary_loss = tf.summary.scalar("Loss", self.loss)
        self.reward = tf.placeholder("float", name="reward")
        summary_rewards = tf.summary.scalar("Reward", self.reward)
        self.episode_length = tf.placeholder("float", name="episode_length")
        summary_episode_lengths = tf.summary.scalar("Episode_length", self.episode_length)
        self.summary_op = tf.summary.merge([summary_loss, summary_rewards, summary_episode_lengths])

        self.jobs = []
        for i, env in enumerate(self.envs):
            self.jobs.append(
                self.make_thread(
                    env,
                    i,
                    self.config["switch_at_iter"] if self.config["switch_at_iter"] is not None and i != len(self.envs) - 1 else self.config["n_iter"],
                    start_at_iter=(0 if self.config["switch_at_iter"] is None or i != len(self.envs) - 1 else self.config["switch_at_iter"])))

        for i, job in enumerate(self.jobs):
            only_sparse = (self.config["switch_at_iter"] is not None and i == len(self.jobs) - 1)
            grads = tf.gradients(job.loss, (self.shared_vars if not(only_sparse) else []) + [job.sparse_representation])
            job.apply_grad = job.optimizer.apply_gradients(
                zip(
                    grads,
                    (self.shared_vars if not(only_sparse) else []) + [job.sparse_representation]
                ),
                global_step=self.global_step
            )

        self.session.run(tf.global_variables_initializer())

        if self.config["save_model"]:
            for job in self.jobs:
                tf.add_to_collection("action", job.action)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()

    def build_networks(self):
        with tf.variable_scope("shared"):
            self.states = tf.placeholder(tf.float32, [None] + list(self.envs[0].observation_space.shape), name="states")
            self.action_taken = tf.placeholder(tf.float32, name="action_taken")
            self.advantage = tf.placeholder(tf.float32, name="advantage")

            if self.config["feature_extraction"]:
                self.L1 = tf.contrib.layers.fully_connected(
                    inputs=self.states,
                    num_outputs=self.config["n_hidden_units"],
                    activation_fn=tf.tanh,
                    weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
                    biases_initializer=tf.zeros_initializer(),
                    scope="L1")
            else:
                self.L1 = self.states
            self.knowledge_base = tf.Variable(tf.truncated_normal([self.L1.get_shape()[-1].value, self.config["n_sparse_units"]], mean=0.0, stddev=0.02), name="knowledge_base")

            self.shared_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

    def signal_handler(self, signal, frame):
        """When a (SIGINT) signal is received, request the threads (via the master) to stop after completing an iteration."""
        logging.info("SIGINT signal received: Requesting a stop...")
        self.stop_requested = True

    def learn(self):
        signal.signal(signal.SIGINT, self.signal_handler)
        if self.config["switch_at_iter"] is None:
            idx = None
        else:
            idx = -1
        for job in self.jobs[:idx]:
            job.start()
        for job in self.jobs[:idx]:
            job.join()
        try:
            self.jobs[idx].start()
            self.jobs[idx].join()
        except TypeError:  # idx is None
            pass

        if self.config["save_model"]:
            self.saver.save(self.session, os.path.join(self.monitor_path, "model"))

    def make_thread(self, env, task_id, n_iter, start_at_iter=0):
        return AKTThread(self, env, task_id, n_iter, start_at_iter=start_at_iter)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import argparse
import tensorflow as tf
from gym import wrappers

from yarll.environment.registration import make

class ModelRunner(object):
    """
    Run an already learned model.
    Currently only supports one variation of an environment.
    """
    def __init__(self, env, model_directory: str, save_directory: str, **usercfg) -> None:
        super(ModelRunner, self).__init__()
        self.env = env
        self.model_directory = model_directory
        self.save_directory = save_directory
        self.config = dict(
            episode_max_length=self.env.spec.tags.get('wrapper_config.TimeLimit.max_episode_steps'),
            repeat_n_actions=1
        )
        self.config.update(usercfg)

        self.session = tf.Session()
        self.saver = tf.train.import_meta_graph(os.path.join(self.model_directory, "model.meta"))
        self.saver.restore(self.session, os.path.join(self.model_directory, "model"))

        self.action = tf.get_collection("action")[0]
        self.states = tf.get_collection("states")[0]

    def choose_action(self, state):
        """Choose an action."""
        return self.session.run([self.action], feed_dict={self.states: [state]})[0]

    def get_trajectory(self, render: bool = False):
        """
        Run agent-environment loop for one whole episode (trajectory)
        Return dictionary of results
        """
        state = self.env.reset()
        for _ in range(self.config["episode_max_length"]):
            action = self.choose_action(state)
            for _ in range(self.config["repeat_n_actions"]):
                _, _, done, _ = self.env.step(action)
                if done:  # Don't continue if episode has already ended
                    break
            if done:
                break
            if render:
                self.env.render()
        return

    def run(self):
        for _ in range(self.config["n_iter"]):
            self.get_trajectory()


parser = argparse.ArgumentParser()
parser.add_argument("environment", metavar="env", type=str, help="Gym environment to execute the model on.")
parser.add_argument("model_directory", type=str, help="Directory from where model files are loaded.")
parser.add_argument("save_directory", type=str, help="Directory where results of running the model are saved")
parser.add_argument("--iterations", default=100, type=int, help="Number of iterations to run the algorithm.")

def main():
    args = parser.parse_args()
    env = make(args.environment)
    runner = ModelRunner(env, args.model_directory, args.save_directory, n_iter=args.iterations)
    try:
        runner.env = wrappers.Monitor(runner.env, args.save_directory, video_callable=False, force=True)
        runner.run()
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()



# -*- coding: utf8 -*-

#  Policy Gradient Implementation
#  Adapted for Tensorflow
#  Other differences:
#  - Always choose the action with the highest probability
#  Source: http://rl-gym-doc.s3-website-us-west-2.amazonaws.com/mlss/lab2.html

import os
from typing import Dict
import numpy as np

import tensorflow as tf
from gym import wrappers

from yarll.agents.agent import Agent
from yarll.agents.env_runner import EnvRunner
from yarll.misc.utils import discount_rewards, flatten, FastSaver
from yarll.misc.network_ops import conv2d, linear, normalized_columns_initializer
from yarll.misc.reporter import Reporter

class REINFORCE(Agent):
    """
    REINFORCE with baselines
    """
    def __init__(self, env, monitor_path: str, monitor: bool = False, video: bool = True, **usercfg) -> None:
        super(REINFORCE, self).__init__(**usercfg)
        self.env = env
        if monitor:
            self.env = wrappers.Monitor(self.env,
                                        monitor_path,
                                        force=True,
                                        video_callable=(None if video else False))
        self.monitor_path = monitor_path
        # Default configuration. Can be overwritten using keyword arguments.
        self.config.update(dict(
            batch_update="timesteps",
            timesteps_per_batch=1000,
            n_iter=100,
            gamma=0.99,  # Discount past rewards by a percentage
            learning_rate=0.05,
            entropy_coef=1e-3,
            n_hidden_layers=2,
            n_hidden_units=20,
            repeat_n_actions=1,
            save_model=False
        ))
        self.config.update(usercfg)

        self.states = tf.placeholder(
            tf.float32, [None] + list(self.env.observation_space.shape), name="states")  # Observation
        self.actions_taken = tf.placeholder(tf.float32, name="actions_taken")  # Discrete action
        self.advantage = tf.placeholder(tf.float32, name="advantage")  # Advantage
        self.build_network()
        self.make_trainer()

        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()
        summary_loss = tf.summary.scalar("model/loss", self.summary_loss)
        summary_entropy = tf.summary.scalar("model/entropy", self.entropy)
        self.summary_op = tf.summary.merge([summary_loss, summary_entropy])

        self.init_op = tf.global_variables_initializer()
        # Launch the graph.
        self.session = tf.Session()
        self.writer = tf.summary.FileWriter(os.path.join(self.monitor_path, "task0"), self.session.graph)

        self.env_runner = EnvRunner(self.env, self, usercfg, summary_writer=self.writer)

    def _initialize(self) -> None:
        self.session.run(self.init_op)

    def build_network(self):
        raise NotImplementedError()

    def make_trainer(self):
        raise NotImplementedError()

    def choose_action(self, state, features) -> Dict[str, np.ndarray]:
        """Choose an action."""
        action = self.session.run([self.action], feed_dict={self.states: [state]})[0]
        return {"action": action}

    def learn(self):
        """Run learning algorithm"""
        self._initialize()
        reporter = Reporter()
        config = self.config
        total_n_trajectories = 0
        for iteration in range(config["n_iter"]):
            # Collect trajectories until we get timesteps_per_batch total timesteps
            trajectories = self.env_runner.get_trajectories()
            total_n_trajectories += len(trajectories)
            all_state = np.concatenate([trajectory.states for trajectory in trajectories])
            # Compute discounted sums of rewards
            rets = [discount_rewards(trajectory.rewards, config["gamma"]) for trajectory in trajectories]
            max_len = max(len(ret) for ret in rets)
            padded_rets = [np.concatenate([ret, np.zeros(max_len - len(ret))]) for ret in rets]
            # Compute time-dependent baseline
            baseline = np.mean(padded_rets, axis=0)
            # Compute advantage function
            advs = [ret - baseline[:len(ret)] for ret in rets]
            all_action = np.concatenate([trajectory.actions for trajectory in trajectories])
            all_adv = np.concatenate(advs)
            # Do policy gradient update step
            episode_rewards = np.array([sum(trajectory.rewards) for trajectory in trajectories])  # episode total rewards
            episode_lengths = np.array([len(trajectory.rewards) for trajectory in trajectories])  # episode lengths
            # TODO: deal with RNN state
            summary, _ = self.session.run([self.summary_op, self.train], feed_dict={
                self.states: all_state,
                self.actions_taken: all_action,
                self.advantage: all_adv
            })
            self.writer.add_summary(summary, iteration)
            self.writer.flush()

            reporter.print_iteration_stats(iteration, episode_rewards, episode_lengths, total_n_trajectories)
        if self.config["save_model"]:
            self.saver.save(self.session, os.path.join(self.monitor_path, "model"))

class REINFORCEDiscrete(REINFORCE):
    def __init__(self, env, monitor_path: str, video: bool = True, **usercfg) -> None:
        super(REINFORCEDiscrete, self).__init__(env, monitor_path, video=video, **usercfg)

    def make_trainer(self):
        good_probabilities = tf.reduce_sum(tf.multiply(self.probs,
                                                       tf.one_hot(tf.cast(self.actions_taken, tf.int32),
                                                                  self.env_runner.nA)),
                                           reduction_indices=[1])
        eligibility = tf.log(good_probabilities) * self.advantage
        loss = -tf.reduce_sum(eligibility)
        self.summary_loss = loss
        optimizer = tf.train.AdamOptimizer(learning_rate=self.config["learning_rate"])
        self.train = optimizer.minimize(loss)

    def build_network(self):

        L1 = tf.contrib.layers.fully_connected(
            inputs=self.states,
            num_outputs=int(self.config["n_hidden_units"]),
            activation_fn=tf.tanh,
            weights_initializer=tf.random_normal_initializer(),
            biases_initializer=tf.zeros_initializer())

        self.probs = tf.contrib.layers.fully_connected(
            inputs=L1,
            num_outputs=self.env_runner.nA,
            activation_fn=tf.nn.softmax,
            weights_initializer=tf.random_normal_initializer(),
            biases_initializer=tf.zeros_initializer())

        self.action = tf.squeeze(tf.multinomial(tf.log(self.probs), 1), name="action")

class REINFORCEDiscreteCNN(REINFORCEDiscrete):
    def __init__(self, env, monitor_path, video=True, **usercfg):
        usercfg["n_hidden_units"] = 200
        super(REINFORCEDiscreteCNN, self).__init__(env, monitor_path, video=video, **usercfg)
        self.config.update(usercfg)

    def build_network(self):
        shape = list(self.env.observation_space.shape)

        x = self.states
        # Convolution layers
        for i in range(4):
            x = tf.nn.elu(conv2d(x, 32, "l{}".format(i + 1), [3, 3], [2, 2]))

        # Flatten
        shape = x.get_shape().as_list()
        reshape = tf.reshape(x, [-1, shape[1] * shape[2] * shape[3]])  # -1 for the (unknown) batch size

        # Fully connected layer 1
        self.L3 = tf.contrib.layers.fully_connected(
            inputs=reshape,
            num_outputs=int(self.config["n_hidden_units"]),
            activation_fn=tf.nn.relu,
            weights_initializer=tf.random_normal_initializer(stddev=0.01),
            biases_initializer=tf.zeros_initializer())

        # Fully connected layer 2
        self.probs = tf.contrib.layers.fully_connected(
            inputs=self.L3,
            num_outputs=self.env_runner.nA,
            activation_fn=tf.nn.softmax,
            weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
            biases_initializer=tf.zeros_initializer())

        self.action = tf.squeeze(tf.multinomial(tf.log(self.probs), 1), name="action")

class REINFORCEDiscreteRNN(REINFORCEDiscrete):
    def __init__(self, env, monitor_path, video=True, **usercfg):
        super(REINFORCEDiscreteRNN, self).__init__(env, monitor_path, video=video, **usercfg)

    def build_network(self):

        n_states = tf.shape(self.states)[:1]
        states = tf.expand_dims(flatten(self.states), [0])

        enc_cell = tf.contrib.rnn.GRUCell(int(self.config["n_hidden_units"]))
        self.rnn_state_in = enc_cell.zero_state(1, tf.float32)
        L1, self.rnn_state_out = tf.nn.dynamic_rnn(cell=enc_cell,
                                                   inputs=states,
                                                   sequence_length=n_states,
                                                   initial_state=self.rnn_state_in,
                                                   dtype=tf.float32)
        self.probs = tf.contrib.layers.fully_connected(
            inputs=L1[0],
            num_outputs=self.env_runner.nA,
            activation_fn=tf.nn.softmax,
            weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
            biases_initializer=tf.zeros_initializer())
        self.action = tf.squeeze(tf.multinomial(tf.log(self.probs), 1), name="action")

    def choose_action(self, state, features):
        """Choose an action."""
        feed_dict = {
            self.states: [state],
            self.rnn_state_in: features
        }
        action, new_features = self.session.run([self.action, self.rnn_state_out], feed_dict=feed_dict)
        return {"action": action, "features": new_features}

class REINFORCEDiscreteCNNRNN(REINFORCEDiscreteRNN):
    def __init__(self, env, monitor_path, video=True, **usercfg):
        super(REINFORCEDiscreteCNNRNN, self).__init__(env, monitor_path, video=video, **usercfg)

    def build_network(self):

        shape = list(self.env.observation_space.shape)

        self.states = tf.placeholder(tf.float32, [None] + shape, name="states")
        self.N = tf.placeholder(tf.int32, name="N")

        x = self.states
        # Convolution layers
        for i in range(4):
            x = tf.nn.elu(conv2d(x, 32, "l{}".format(i + 1), [3, 3], [2, 2]))

        # Flatten
        shape = x.get_shape().as_list()
        reshape = tf.reshape(x, [-1, shape[1] * shape[2] * shape[3]])  # -1 for the (unknown) batch size

        reshape = tf.expand_dims(flatten(reshape), [0])
        self.enc_cell = tf.contrib.rnn.BasicLSTMCell(int(self.config["n_hidden_units"]))
        self.rnn_state_in = self.enc_cell.zero_state(1, tf.float32)
        self.L3, self.rnn_state_out = tf.nn.dynamic_rnn(cell=self.enc_cell,
                                                        inputs=reshape,
                                                        initial_state=self.rnn_state_in,
                                                        dtype=tf.float32)

        self.probs = tf.contrib.layers.fully_connected(
            inputs=self.L3[0],
            num_outputs=self.env_runner.nA,
            activation_fn=tf.nn.softmax,
            weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
            biases_initializer=tf.zeros_initializer())
        self.action = tf.squeeze(tf.multinomial(tf.log(self.probs), 1), name="action")

class REINFORCEContinuous(REINFORCE):
    def __init__(self, env, monitor_path, RNN=False, video=True, **usercfg):
        self.rnn = RNN
        super(REINFORCEContinuous, self).__init__(env, monitor_path, video=video, **usercfg)

    def build_network(self):
        if self.rnn:
            self.build_network_rnn()
        else:
            self.build_network_normal()

    def make_trainer(self):
        loss = tf.reduce_mean(-self.action_log_prob * self.advantage) - self.config["entropy_coef"] * self.entropy
        self.summary_loss = tf.reduce_mean(loss)
        optimizer = tf.train.AdamOptimizer(learning_rate=self.config["learning_rate"])
        self.train = optimizer.minimize(loss)

    def build_network_normal(self):

        self.actions_taken = tf.placeholder(
            tf.float32,
            [None] + list(self.env.action_space.shape),
            name="actions_taken")

        x = self.states
        for i in range(int(self.config["n_hidden_layers"])):
            x = tf.tanh(linear(x, int(self.config["n_hidden_units"]), "L{}_mean".format(i + 1),
                               initializer=normalized_columns_initializer(1.0)))
        self.mean = linear(x, self.env.action_space.shape[0], "mean", initializer=normalized_columns_initializer(0.01))
        self.mean = tf.check_numerics(self.mean, "mean")

        self.log_std = tf.get_variable(
            name="logstd",
            shape=list(self.env.action_space.shape),
            initializer=tf.zeros_initializer()
        )
        std = tf.exp(self.log_std, name="std")
        std = tf.check_numerics(std, "std")

        self.action = self.mean + std * tf.random_normal(tf.shape(self.mean))
        self.action = tf.reshape(self.action, list(self.env.action_space.shape))

        neglogprob = 0.5 * tf.reduce_sum(tf.square((self.actions_taken - self.mean) / std), axis=-1) \
            + 0.5 * np.log(2.0 * np.pi) * tf.to_float(tf.shape(self.actions_taken)[-1]) \
            + tf.reduce_sum(self.log_std, axis=-1)
        self.action_log_prob = -neglogprob
        self.entropy = tf.reduce_sum(self.log_std + .5 * np.log(2.0 * np.pi * np.e), axis=-1)

    def build_network_rnn(self):
        n_states = tf.shape(self.states)[:1]
        states = tf.expand_dims(flatten(self.states), [0])

        enc_cell = tf.contrib.rnn.GRUCell(int(self.config["n_hidden_units"]))
        L1, _ = tf.nn.dynamic_rnn(cell=enc_cell, inputs=states,
                                  sequence_length=n_states, dtype=tf.float32)

        L1 = L1[0]

        mu, sigma = mu_sigma_layer(L1, 1)

        self.normal_dist = tf.contrib.distributions.Normal(mu, sigma)
        self.action = self.normal_dist.sample(1)
        self.action = tf.clip_by_value(self.action, self.env.action_space.low[0], self.env.action_space.high[0])



# -*- coding: utf8 -*-

import gym
import numpy as np


class DescriptionWrapper(gym.Wrapper):
    """
    Wrapper that mantains a description of the environment
    in its metadata.
    """

    changeable_parameters: list = []

    def __init__(self, env, **kwargs):
        super(DescriptionWrapper, self).__init__(env)

        self.args: dict = kwargs
        self.metadata = self.metadata.copy()
        if "changeable_parameters" not in self.metadata:
            self.metadata["changeable_parameters"] = self.changeable_parameters
        self.metadata["parameters"] = {"env_id": self.spec.id}
        self.metadata["parameters"].update(self.changeable_parameters_values())

    def changeable_parameters_values(self) -> dict:
        params: dict = {}
        for p in self.metadata["changeable_parameters"]:
            params[p["name"]] = self.unwrapped.__getattribute__(p["name"])
        return params


class DiscreteObservationWrapper(gym.ObservationWrapper):
    """
    Converts envs with a discrete observation space (Discrete(N)) to one
    with a one-hot encoding (Box(N,)).
    """

    def __init__(self, env):
        super(DiscreteObservationWrapper, self).__init__(env)

        if not isinstance(self.env.observation_space, gym.spaces.Discrete):
            raise AssertionError(
                "This wrapper can only be applied to environments with a discrete observation space.")
        self.n: int = self.observation_space.n
        self.observation_space: gym.spaces.Box = gym.spaces.Box(
            low=0.0,
            high=1.0,
            shape=(self.env.observation_space.n,))

    def observation(self, observation: int) -> np.ndarray:
        converted = np.zeros(self.n)
        converted[observation] = 1.0
        return converted


class NormalizedObservationWrapper(gym.ObservationWrapper):
    """
    Normalizes observations such that the values are
    between 0.0 and 1.0.
    """

    def __init__(self, env):
        super(NormalizedObservationWrapper, self).__init__(env)
        if not isinstance(self.env.observation_space, gym.spaces.Box):
            raise AssertionError(
                "This wrapper can only be applied to environments with a continuous observation space.")
        if np.inf in self.env.observation_space.low or np.inf in self.env.observation_space.high:
            raise AssertionError(
                "This wrapper cannot be used for observation spaces with an infinite lower/upper bound.")
        self.observation_space: gym.spaces.Box = gym.spaces.Box(
            low=np.zeros(self.env.observation_space.shape),
            high=np.ones(self.env.observation_space.shape)
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return (observation - self.env.observation_space.low) / \
            (self.env.observation_space.high - self.env.observation_space.low)


class NormalizedRewardWrapper(gym.RewardWrapper):
    """
    Normalizes rewards such that the values are between 0.0 and 1.0.
    """

    def __init__(self, env, low, high):
        super(NormalizedRewardWrapper, self).__init__(env)
        self.low = low
        self.high = high
        self.reward_range = (0.0, 1.0)

    def reward(self, rew):
        return (rew - self.low) / (self.high - self.low)



import gym
from yarll.environment.registration import EnvSpec

def register_env(name, entry_point: str, tags=None, **kwargs):
    already_registered = name in gym.envs.registry.env_specs
    if already_registered:
        old_env_name = "Old" + name
        gym.envs.registry.env_specs[old_env_name] = gym.envs.registry.env_specs[name]
        new_tags = gym.envs.registry.env_specs[name].tags
    else:
        new_tags = {}
    if tags is not None:
        new_tags.update(tags)
    gym.envs.registry.env_specs[name] = EnvSpec(
        name,
        entry_point=entry_point,
        kwargs={"old_env_name": old_env_name} if already_registered else {},
        tags=new_tags,
        **kwargs)

register_env(
    "CartPole-v0",
    entry_point="yarll.environment.cartpole:CartPole",
    max_episode_steps=200
)

register_env(
    "Acrobot-v1",
    entry_point="yarll.environment.acrobot:Acrobot",
    max_episode_steps=500
)

register_env(
    "FrozenLake8x8-v0",
    entry_point="yarll.environment.environment:Environment",
    tags={
        "wrappers": ["environment.wrappers:DiscreteObservationWrapper"]
    }
)



# -*- coding: utf8 -*-

import logging
import matplotlib
import numpy as np

gui_env = ['TKAgg', 'GTKAgg', 'Qt4Agg', 'WXAgg', 'agg']
for gui in gui_env:
    try:
        matplotlib.use(gui, warn=False, force=True)
        from matplotlib import pyplot as plt
        break
    except ImportError:
        continue

class Reporter(object):
    """Report iteration statistics using text and graphs."""
    def __init__(self):
        super(Reporter, self).__init__()
        logging.getLogger().setLevel("INFO")
        self.fig = None
        self.ax1 = None

    @staticmethod
    def print_iteration_stats(iteration: int, episode_rewards: np.ndarray, episode_lengths: np.ndarray, total_n_trajectories: int):
        """Print statistics about rewards and episode lengths of the current iteration."""
        logging.info("-----------------")
        logging.info("Iteration: \t\t {}".format(iteration))
        logging.info("NumTrajs: \t\t {}".format(len(episode_rewards)))
        logging.info("NumTimesteps: \t {}".format(np.sum(episode_lengths)))
        logging.info("MaxRew: \t\t {}".format(episode_rewards.max()))
        logging.info("MeanRew: \t\t {} +- {}".format(episode_rewards.mean(), episode_rewards.std() / np.sqrt(len(episode_rewards))))
        logging.info("MeanLen: \t\t {} +- {}".format(episode_lengths.mean(), episode_lengths.std() / np.sqrt(len(episode_lengths))))
        logging.info("Total NumTrajs: \t {}".format(total_n_trajectories))
        logging.info("-----------------")
        return

    def draw_rewards(self, mean_rewards):
        """Draw a plot with the mean reward for each batch of episodes."""
        if not self.fig:
            self.fig = plt.figure()
        if not self.ax1:
            self.ax1 = self.fig.add_subplot(1, 1, 1)
        self.ax1.clear()
        self.ax1.plot(range(len(mean_rewards)), mean_rewards)
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()
        plt.show(block=False)



# -*- coding: utf8 -*-

from typing import Optional
import gym

from yarll.environment.environment import Environment

class Atari(Environment):
    """Atari environment wrapper."""

    changeable_parameters = [  # TODO: use dict instead of list of dicts
        {
            "name": "difficulty",
            "type": "choice",
            "options": []  # Depends on the environment; updated at runtime
        }
    ]

    def __init__(self, old_env_name: str, difficulty: Optional[int] = None, **kwargs) -> None:
        super(Atari, self).__init__(gym.make(old_env_name), extra_attrs={"difficulty": difficulty}, **kwargs)
        self.changeable_parameters[0]["options"] = self.unwrapped.ale.getAvailableDifficulties()
        self.unwrapped.difficulty = difficulty
        self.change_parameters(difficulty=difficulty)

    def change_parameters(self, difficulty: Optional[int] = None):
        """Change an Atari environment using a different difficulty."""
        if difficulty is not None:
            self.unwrapped.difficulty = difficulty
            self.unwrapped.ale.setDifficulty(difficulty)



# -*- coding: utf8 -*-

"""
Functions and networks for actor-critic agents.
"""

from typing import Sequence
import tensorflow as tf
import numpy as np

from yarll.misc.network_ops import normalized_columns_initializer, linear, conv2d, flatten

class ActorCriticNetwork(object):
    pass

class ActorCriticNetworkDiscrete(ActorCriticNetwork):
    """
    Neural network for the Actor of an Actor-Critic algorithm using a discrete action space.
    """

    def __init__(self, state_shape: Sequence[int], n_actions: int, n_hidden_units: int, n_hidden_layers: int) -> None:
        super(ActorCriticNetworkDiscrete, self).__init__()
        self.state_shape = state_shape
        self.n_actions = n_actions

        self.states = tf.placeholder(tf.float32, [None] + list(state_shape), name="states")
        self.actions_taken = tf.placeholder(tf.float32, [None, n_actions], name="actions_taken")

        x = self.states
        for i in range(n_hidden_layers):
            x = tf.tanh(linear(x, n_hidden_units, "L{}_action".format(i + 1),
                               initializer=normalized_columns_initializer(1.0)))
        self.logits = linear(x, n_actions, "actionlogits", normalized_columns_initializer(0.01))

        x = self.states
        for i in range(n_hidden_layers):
            x = tf.tanh(linear(x, n_hidden_units, "L{}_value".format(i + 1),
                               initializer=normalized_columns_initializer(1.0)))
        self.value = tf.reshape(linear(x, 1, "value", normalized_columns_initializer(1.0)), [-1])

        self.probs = tf.nn.softmax(self.logits)

        self.action = tf.squeeze(tf.multinomial(
            self.logits - tf.reduce_max(self.logits, [1], keepdims=True), 1), [1], name="action")
        self.action = tf.one_hot(self.action, n_actions)[0, :]

        # Log probabilities of all actions
        self.log_probs = tf.nn.log_softmax(self.logits)
        # Prob of the action that was actually taken
        self.action_log_prob = tf.reduce_sum(self.log_probs * self.actions_taken, [1])

        self.entropy = self.probs * self.log_probs


class ActorCriticNetworkDiscreteCNN(ActorCriticNetwork):
    """docstring for ActorCriticNetworkDiscreteCNNRNN"""

    def __init__(self, state_shape: Sequence[int], n_actions: int, n_hidden: int, summary: bool = True) -> None:
        super(ActorCriticNetworkDiscreteCNN, self).__init__()
        self.state_shape = state_shape
        self.n_actions = n_actions
        self.n_hidden = n_hidden
        self.summary = summary

        self.states = tf.placeholder(tf.float32, [None] + state_shape, name="states")
        self.actions_taken = tf.placeholder(tf.float32, [None, n_actions], name="actions_taken")

        x = self.states
        # Convolution layers
        for i in range(4):
            x = tf.nn.elu(conv2d(x, 32, "l{}".format(i + 1), [3, 3], [2, 2]))

        # Flatten
        shape = x.get_shape().as_list()
        # -1 for the (unknown) batch size
        reshape = tf.reshape(x, [-1, shape[1] * shape[2] * shape[3]])

        # Fully connected for actor & critic
        self.logits = linear(reshape, n_actions, "actionlogits", normalized_columns_initializer(0.01))
        self.value = tf.reshape(
            linear(reshape, 1, "value", normalized_columns_initializer(1.0)), [-1])

        self.probs = tf.nn.softmax(self.logits)

        self.action = tf.squeeze(tf.multinomial(
            self.logits - tf.reduce_max(self.logits, [1], keepdims=True), 1), [1], name="action")
        self.action = tf.one_hot(self.action, n_actions)[0, :]

        # Log probabilities of all actions
        self.log_probs = tf.nn.log_softmax(self.logits)
        # Prob of the action that was actually taken
        self.action_log_prob = tf.reduce_sum(self.log_probs * self.actions_taken, [1])

        self.entropy = self.probs * self.log_probs


class ActorCriticNetworkDiscreteCNNRNN(ActorCriticNetwork):
    """docstring for ActorCriticNetworkDiscreteCNNRNN"""

    def __init__(self,
                 state_shape: Sequence[int],
                 n_actions: int,
                 n_hidden: int,
                 lstm_size: int = 256,
                 summary: bool = True) -> None:
        super(ActorCriticNetworkDiscreteCNNRNN, self).__init__()
        self.state_shape: Sequence[int] = state_shape
        self.n_actions: int = n_actions
        self.n_hidden: int = n_hidden
        self.summary: bool = summary

        self.states = tf.placeholder(tf.float32, [None] + state_shape, name="states")
        self.actions_taken = tf.placeholder(tf.float32, name="actions_taken")

        x = self.states
        # Convolution layers
        for i in range(4):
            x = tf.nn.elu(conv2d(x, 32, "l{}".format(i + 1), [3, 3], [2, 2]))

        # Flatten
        reshape = tf.expand_dims(flatten(x), [0])

        self.enc_cell = tf.contrib.rnn.BasicLSTMCell(lstm_size)
        lstm_state_size = self.enc_cell.state_size
        c_init = np.zeros((1, lstm_state_size.c), np.float32)
        h_init = np.zeros((1, lstm_state_size.h), np.float32)
        self.state_init = [c_init, h_init]
        self.rnn_state_in = self.enc_cell.zero_state(1, tf.float32)
        tf.add_to_collection("rnn_state_in_c", self.rnn_state_in.c)
        tf.add_to_collection("rnn_state_in_h", self.rnn_state_in.h)
        L3, self.rnn_state_out = tf.nn.dynamic_rnn(cell=self.enc_cell,
                                                   inputs=reshape,
                                                   initial_state=self.rnn_state_in,
                                                   dtype=tf.float32)
        tf.add_to_collection("rnn_state_out_c", self.rnn_state_out.c)
        tf.add_to_collection("rnn_state_out_h", self.rnn_state_out.h)
        L3 = tf.reshape(L3, [-1, lstm_size])

        # Fully connected for actor and critic
        self.logits = linear(L3, n_actions, "actionlogits", normalized_columns_initializer(0.01))
        self.value = tf.reshape(linear(L3, 1, "value", normalized_columns_initializer(1.0)), [-1])

        self.probs = tf.nn.softmax(self.logits)

        self.action = tf.squeeze(tf.multinomial(
            self.logits - tf.reduce_max(self.logits, [1], keepdims=True), 1), [1], name="action")
        self.action = tf.one_hot(self.action, n_actions)[0, :]

        # Log probabilities of all actions
        self.log_probs = tf.nn.log_softmax(self.logits)
        # Prob of the action that was actually taken
        self.action_log_prob = tf.reduce_sum(self.log_probs * self.actions_taken, [1])

        self.entropy = self.probs * self.log_probs


def actor_critic_discrete_loss(logits,
                               probs,
                               value,
                               actions_taken,
                               advantage,
                               ret,
                               vf_coef: float = 0.5,
                               entropy_coef: float = 0.01,
                               reducer="sum"):
    tf_reducer = tf.reduce_sum if reducer == "sum" else tf.reduce_mean
    log_probs = tf.nn.log_softmax(logits)
    actor_loss = - tf_reducer(tf.reduce_sum(log_probs * actions_taken, [1]) * advantage)
    critic_loss = tf_reducer(tf.square(value - ret))
    entropy = tf_reducer(probs * log_probs)
    loss = actor_loss + vf_coef * critic_loss - entropy_coef * entropy
    return actor_loss, critic_loss, loss


class ActorCriticNetworkContinuous(ActorCriticNetwork):
    """Neural network for an Actor of an Actor-Critic algorithm using a continuous action space."""

    def __init__(self, state_shape: Sequence[int], action_space, n_hidden_units: int, n_hidden_layers: int = 1) -> None:
        super(ActorCriticNetworkContinuous, self).__init__()
        self.state_shape = state_shape

        self.states = tf.placeholder("float", [None] + list(state_shape), name="states")
        self.actions_taken = tf.placeholder(
            tf.float32,
            [None] + list(action_space.shape),
            name="actions_taken")

        x = self.states
        for i in range(n_hidden_layers):
            x = tf.tanh(linear(x, n_hidden_units, "L{}_mean".format(i + 1),
                               initializer=normalized_columns_initializer(1.0)))
        self.mean = linear(x, action_space.shape[0], "mean", initializer=normalized_columns_initializer(0.01))
        self.mean = tf.check_numerics(self.mean, "mean")

        self.log_std = tf.get_variable(
            name="logstd",
            shape=list(action_space.shape),
            initializer=tf.zeros_initializer()
        )
        std = tf.exp(self.log_std, name="std")
        std = tf.check_numerics(std, "std")

        self.action = self.mean + std * tf.random_normal(tf.shape(self.mean))
        self.action = tf.reshape(self.action, list(action_space.shape))

        x = self.states
        for i in range(n_hidden_layers):
            x = tf.tanh(linear(x, n_hidden_units, "L{}_value".format(i + 1),
                               initializer=normalized_columns_initializer(1.0)))

        self.value = tf.reshape(linear(x, 1, "value", normalized_columns_initializer(1.0)), [-1])

        neglogprob = 0.5 * tf.reduce_sum(tf.square((self.actions_taken - self.mean) / std), axis=-1) \
            + 0.5 * np.log(2.0 * np.pi) * tf.to_float(tf.shape(self.actions_taken)[-1]) \
            + tf.reduce_sum(self.log_std, axis=-1)
        self.action_log_prob = -neglogprob
        self.entropy = -tf.reduce_sum(self.log_std + .5 * np.log(2.0 * np.pi * np.e), axis=-1)


def actor_critic_continuous_loss(action_log_prob,
                                 entropy,
                                 value,
                                 advantage,
                                 ret,
                                 vf_coef: float = 0.5,
                                 entropy_coef: float = 0.01,
                                 reducer: str = "sum"):
    tf_reducer = tf.reduce_sum if reducer == "sum" else tf.reduce_mean
    actor_loss = - tf_reducer(action_log_prob * advantage)
    critic_loss = tf_reducer(tf.square(value - ret))
    loss = actor_loss + vf_coef * critic_loss - entropy_coef * entropy
    return actor_loss, critic_loss, loss



# -*- coding: utf8 -*-

"""
Asynchronous Advantage Actor Critic (A3C)
Most of the work is done in `a3c_worker.py`.
Based on:
- Pseudo code from Asynchronous Methods for Deep Reinforcement Learning
- Tensorflow code from https://github.com/yao62995/A3C/blob/master/A3C_atari.py and
  https://github.com/openai/universe-starter-agent/tree/f16f37d9d3bc8146cf68a75557e1ba89824b7e54
"""

import logging
import multiprocessing
import subprocess
import signal
import sys
import os
from typing import Optional
from six.moves import shlex_quote

from yarll.agents.agent import Agent

logging.getLogger().setLevel("INFO")

class A3C(Agent):
    """Asynchronous Advantage Actor Critic learner."""

    def __init__(self, env, monitor_path: str, monitor: bool = False, video: bool = True, **usercfg) -> None:
        super(A3C, self).__init__(**usercfg)
        self.env = env
        self.env_name = env.spec.id
        self.monitor = monitor
        self.monitor_path = monitor_path
        self.video = video
        self.task_type: Optional[str] = None # To be filled in by subclass

        self.config.update(dict(
            gamma=0.99,  # Discount past rewards by a percentage
            learning_rate=1e-4,
            n_hidden_units=20,
            n_hidden_layers=1,
            gradient_clip_value=50,
            n_tasks=multiprocessing.cpu_count(),  # Use as much tasks as there are cores on the current system
            T_max=8e5,
            shared_optimizer=False,
            episode_max_length=env.spec.tags.get("wrapper_config.TimeLimit.max_episode_steps"),
            n_local_steps=20,
            vf_coef=0.5,
            entropy_coef=0.01,
            loss_reducer="sum",  # use tf.reduce_sum or tf.reduce_mean for the loss
            save_model=False
        ))
        self.config.update(usercfg)

        self.current_folder = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
        self.ps_process: Optional[subprocess.Popen] = None

    def signal_handler(self, received_signal: int, frame):
        logging.info("SIGINT signal received: Requesting a stop...")
        sys.exit(128 + received_signal)

    def start_parameter_server(self):
        cmd = [
            sys.executable,
            os.path.join(self.current_folder, "parameter_server.py"),
            self.config["n_tasks"]]
        processed_cmd = " ".join(shlex_quote(str(x)) for x in cmd)
        self.ps_process = subprocess.Popen(processed_cmd, shell=True)

    def stop_parameter_server(self):
        self.ps_process.terminate()

    def start_signal_handler(self):
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGHUP, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

    def learn(self):
        self.start_signal_handler()
        self.start_parameter_server()
        worker_processes = []
        for task_id in range(int(self.config["n_tasks"])):
            cmd = [
                sys.executable,
                os.path.join(self.current_folder, "a3c_worker.py"),
                self.env_name,
                self.task_type,
                task_id,
                int(self.config["n_tasks"]),
                self.config["config_path"],
                "--monitor_path", self.monitor_path
            ]
            processed_cmd = " ".join(shlex_quote(str(x)) for x in cmd)
            p = subprocess.Popen(processed_cmd, shell=True)
            worker_processes.append(p)
        for p in worker_processes:
            p.wait()
        self.stop_parameter_server()


class A3CDiscrete(A3C):
    """A3C for a discrete action space"""
    def __init__(self, env, monitor_path: str, monitor: bool = False, **usercfg) -> None:
        super(A3CDiscrete, self).__init__(env, monitor_path, monitor=monitor, **usercfg)
        self.task_type = "A3CTaskDiscrete"

class A3CDiscreteCNN(A3C):
    """A3C for a discrete action space"""
    def __init__(self, env, monitor_path: str, monitor: bool = False, **usercfg) -> None:
        super(A3CDiscreteCNN, self).__init__(env, monitor_path, monitor=monitor, **usercfg)
        self.task_type = "A3CTaskDiscreteCNN"

class A3CDiscreteCNNRNN(A3C):
    """A3C for a discrete action space"""
    def __init__(self, env, monitor_path: str, monitor: bool = False, **usercfg) -> None:
        super(A3CDiscreteCNNRNN, self).__init__(env, monitor_path, monitor=monitor, **usercfg)
        self.task_type = "A3CTaskDiscreteCNNRNN"
        self.config["RNN"] = True

class A3CContinuous(A3C):
    """A3C for a continuous action space"""
    def __init__(self, env, monitor_path: str, monitor: bool = False, **usercfg) -> None:
        super(A3CContinuous, self).__init__(env, monitor_path, monitor=monitor, **usercfg)
        self.task_type = "A3CTaskContinuous"



# -*- coding: utf8 -*-

class Agent(object):
    """Reinforcement learning agent"""
    def __init__(self, **usercfg):
        super(Agent, self).__init__()
        self.config = usercfg
        # Only used (and overwritten) by agents that use an RNN
        self.initial_features = None

    def learn(self):
        """Learn in the current environment."""
        raise NotImplementedError()

    def get_env_action(self, action):
        return action

    def new_trajectory(self):
        """
        Notification by the environment runner that
        a new trajectory will be started
        """

        pass



# -*- coding: utf8 -*-

import numpy as np
import logging

from gym import wrappers

from yarll.agents.agent import Agent
from yarll.misc.utils import discount_rewards
from yarll.misc.reporter import Reporter

logging.getLogger().setLevel("INFO")

np.set_printoptions(suppress=True)  # Don't use the scientific notation to print results

# Adaption of Karpathy's Pong from Pixels article to apply it using a simple neural network on other environments

def scale_state(state, O):
    return state - O.low

def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))

def random_with_probability(output, n_actions, temperature=1.0):
    # total = sum([np.exp(float(o) / temperature) for o in output])
    # probs = [np.exp(float(o) / temperature) / total for o in output]
    probs = output / np.sum(output)
    action = np.random.choice(n_actions, p=probs)
    return action, probs

class Karpathy(Agent):
    """Karpathy policy gradient agent"""
    def __init__(self, env, monitor_path, video=True, **usercfg):
        super(Karpathy, self).__init__(**usercfg)
        self.env = wrappers.Monitor(env, monitor_path, force=True, video_callable=(None if video else False))
        self.nA = self.env.action_space.n
        # Default configuration. Can be overwritten using keyword arguments.
        self.config.update(dict(
            # timesteps_per_batch=10000,
            # n_iter=100,
            episode_max_length=env.spec.tags.get("wrapper_config.TimeLimit.max_episode_steps"),
            gamma=0.99,
            learning_rate=0.05,
            batch_size=10,  # Amount of episodes after which to adapt gradients
            decay_rate=0.99,  # Used for RMSProp
            n_hidden_units=20,
            draw_frequency=50,  # Draw a plot every 50 episodes
            repeat_n_actions=1
        ))
        self.config.update(usercfg)
        self.build_network()

    def build_network(self):
        self.w1 = np.random.randn(self.nO, self.config["n_hidden_units"]) / np.sqrt(self.config["n_hidden_units"])
        self.w2 = np.random.randn(self.config["n_hidden_units"], self.nA) / np.sqrt(self.nA)

    def choose_action(self, state):
        x1, nn_outputs = self.forward_step(state)
        action, probabilities = random_with_probability(nn_outputs, self.nA)
        return action, probabilities, x1

    def forward_step(self, state):
        x1 = np.dot(state, self.w1)
        x1[x1 < 0] = 0  # ReLU
        output = sigmoid(np.dot(x1, self.w2))
        return x1, output

    def backward_step(self, x0, x1, feedback):
        """Computes how much to change the weights from input->hidden layer and hidden->final layer"""
        change_w2 = np.dot(x1.T, feedback)  # 8x200 * 200x3 = 8x3
        dh = np.dot(feedback, self.w2.T)  # 200x3 * 3x8 = 200x8
        dh[x1 <= 0] = 0
        change_w1 = x0.T.dot(dh)  # 2x200 * 200x8 = 2x8
        return change_w1, change_w2

    def get_trajectory(self, render=False):
        """
        Run agent-environment loop for one whole episode (trajectory)
        Return dictionary of results
        Note that this function returns more than the get_trajectory in the Learner class.
        """
        env = self.env
        state = env.reset()
        states = []
        actions = []
        rewards = []
        episode_probabilities = []
        x1s = []
        for _ in range(self.config["episode_max_length"]):
            action, probabilities, x1 = self.choose_action(state)
            x1s.append(x1)
            states.append(state)
            (state, rew, done, _) = env.step(action)
            actions.append(action)
            rewards.append(rew)
            episode_probabilities.append(probabilities)
            if done:
                break
            if render:
                env.render()
        return {"reward": np.array(rewards),
                "state": np.array(states),
                "action": np.array(actions),
                "prob": np.array(episode_probabilities),
                "x1": np.array(x1s)
                }

    def learn(self):
        reporter = Reporter()

        gradient1 = np.zeros_like(self.w1)
        gradient2 = np.zeros_like(self.w2)

        rmsprop1 = np.zeros_like(self.w1)
        rmsprop2 = np.zeros_like(self.w2)

        iteration = 0  # amount of batches processed
        episode_nr = 0
        episode_lengths = np.zeros(self.config["batch_size"])
        episode_rewards = np.zeros(self.config["batch_size"])
        mean_rewards = []
        while True:  # Keep executing episodes
            trajectory = self.get_trajectory(self.config["episode_max_length"])

            episode_rewards[episode_nr % self.config["batch_size"]] = sum(trajectory["reward"])
            episode_lengths[episode_nr % self.config["batch_size"]] = len(trajectory["reward"])
            episode_nr += 1
            action_taken = (np.arange(self.nA) == trajectory["action"][:, None]).astype(np.float32)  # one-hot encoding
            epdlogp = action_taken - trajectory["prob"]

            # episode_states = np.vstack(encountered_states)

            discounted_episode_rewards = discount_rewards(trajectory["reward"], self.config["gamma"])
            # print(discounted_episode_rewards)
            # standardize
            discounted_episode_rewards -= np.mean(discounted_episode_rewards)
            discounted_episode_rewards /= np.std(discounted_episode_rewards)
            epdlogp *= np.reshape(np.repeat(discounted_episode_rewards, self.nA), (len(discounted_episode_rewards), self.nA))

            change_w1, change_w2 = self.backward_step(trajectory["state"], trajectory['x1'], epdlogp)

            gradient1 += change_w1
            gradient2 += change_w2

            if episode_nr % self.config["batch_size"] == 0:  # batch is done
                iteration += 1
                rmsprop1 = self.config["decay_rate"] * rmsprop1 + (1 - self.config["decay_rate"]) * gradient1**2
                rmsprop2 = self.config["decay_rate"] * rmsprop2 + (1 - self.config["decay_rate"]) * gradient2**2
                self.w1 += self.config["learning_rate"] * gradient1 / (np.sqrt(rmsprop1) + 1e-5)
                self.w2 += self.config["learning_rate"] * gradient2 / (np.sqrt(rmsprop2) + 1e-5)
                gradient1 = np.zeros_like(self.w1)
                gradient2 = np.zeros_like(self.w2)
                reporter.print_iteration_stats(iteration, episode_rewards, episode_lengths, episode_nr)
                mean_rewards.append(episode_rewards.mean())
                if episode_nr % self.config["draw_frequency"] == 0:
                    reporter.draw_rewards(mean_rewards)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import argparse
from yarll.misc.utils import json_to_dict, save_config, set_seed

def run_experiment(spec, monitor_path=None, only_last=False, description=None, seed=None):
    """Run an experiment using a specification dictionary."""

    import os

    if seed is not None:
        set_seed(seed)

    import datetime
    import gym
    gym.logger.set_level(gym.logger.ERROR)
    from gym.spaces import Discrete
    from yarll.environment.registration import make, make_environments
    from yarll.agents.registration import make_agent

    args = spec["agent"]["args"]

    if monitor_path:
        args["monitor_path"] = monitor_path
    else:
        monitor_path = args["monitor_path"]
    args["config_path"] = os.path.join(monitor_path, "config.json")
    if not os.path.exists(monitor_path):
        os.makedirs(monitor_path)
    envs_type = spec["environments"]["type"]
    if envs_type == "single":
        envs = [make(spec["environments"]["source"])]
    elif envs_type == "json":
        envs = make_environments(json_to_dict(spec["environments"]["source"]))
    if seed is not None:
        for env in envs:
            env.seed(seed)
    args["seed"] = seed
    args["envs"] = envs
    if len(envs) == 1 or only_last:
        args["env"] = envs[-1]
    action_space_type = "discrete" if isinstance(envs[0].action_space, Discrete) else "continuous"
    state_dimensions = "single" if len(envs[0].observation_space.shape) == 1 else "multi"
    agent = make_agent(spec["agent"]["name"], state_dimensions, action_space_type, **args)
    config = agent.config.copy()
    if description is not None:
        config["description"] = description
    config["seed"] = str(seed)
    config["start_time"] = datetime.datetime.now().astimezone().isoformat()
    save_config(monitor_path, config, [env.metadata["parameters"] for env in envs])
    agent.learn()


parser = argparse.ArgumentParser()
parser.add_argument("experiment", type=str, help="JSON file with the experiment specification.")
parser.add_argument("--description", type=str, help="Experiment description.")
parser.add_argument("--monitor_path", metavar="monitor_path", default=None, type=str,
                    help="Path where Gym monitor files may be saved.")
parser.add_argument("--only_last", default=False, action="store_true",
                    help="Only use the last environment in a list of provided environments.")
parser.add_argument("--seed", default=None, type=int, help="Seed to use for the experiment.")

def main():
    args = parser.parse_args()
    run_experiment(
        json_to_dict(args.experiment),
        monitor_path=args.monitor_path,
        only_last=args.only_last,
        description=args.description,
        seed=args.seed
    )

if __name__ == '__main__':
    main()



from yarll.agents.registration import register_agent, make_agent

register_agent(name="A2C",
               entry_point="yarll.agents.actorcritic.a2c:A2CDiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="A2C",
               entry_point="yarll.agents.actorcritic.a2c:A2CDiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="A2C",
               entry_point="yarll.agents.actorcritic.a2c:A2CDiscreteCNNRNN",
               state_dimensions="multi",
               action_space="discrete",
               RNN=True
              )
register_agent(name="A2C",
               entry_point="yarll.agents.actorcritic.a2c:A2CContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="A3C",
               entry_point="yarll.agents.actorcritic.a3c:A3CDiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="A3C",
               entry_point="yarll.agents.actorcritic.a3c:A3CDiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="A3C",
               entry_point="yarll.agents.actorcritic.a3c:A3CDiscreteCNNRNN",
               state_dimensions="multi",
               action_space="discrete",
               RNN=True
              )
register_agent(name="A3C",
               entry_point="yarll.agents.actorcritic.a3c:A3CContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="AsyncKnowledgeTransfer",
               entry_point="yarll.agents.knowledgetransfer.async_knowledge_transfer:AsyncKnowledgeTransfer",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="PPO",
               entry_point="yarll.agents.ppo.ppo:PPODiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="PPO",
               entry_point="yarll.agents.ppo.ppo:PPODiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="PPO",
               entry_point="yarll.agents.ppo.ppo:PPOContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="DPPO",
               entry_point="yarll.agents.ppo.dppo:DPPODiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="DPPO",
               entry_point="yarll.agents.ppo.dppo:DPPODiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="DPPO",
               entry_point="yarll.agents.ppo.dppo:DPPOContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="TRPO",
               entry_point="yarll.agents.trpo.trpo:TRPODiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="TRPO",
               entry_point="yarll.agents.trpo.trpo:TRPODiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="TRPO",
               entry_point="yarll.agents.trpo.trpo:TRPOContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="DTRPO",
               entry_point="yarll.agents.trpo.dtrpo:DTRPODiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="DTRPO",
               entry_point="yarll.agents.trpo.dtrpo:DTRPODiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="DTRPO",
               entry_point="yarll.agents.trpo.dtrpo:DTRPOContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="DDPG",
               entry_point="yarll.agents.ddpg:DDPG",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="CEM",
               entry_point="yarll.agents.cem:CEM",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="CEM",
               entry_point="yarll.agents.cem:CEM",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="Karpathy",
               entry_point="yarll.agents.karpathy:Karpathy",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="KnowledgeTransfer",
               entry_point="yarll.agents.knowledge_transfer:KnowledgeTransfer",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="REINFORCE",
               entry_point="yarll.agents.reinforce:REINFORCEDiscrete",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="REINFORCE",
               entry_point="yarll.agents.reinforce:REINFORCEDiscreteRNN",
               state_dimensions="single",
               action_space="discrete",
               RNN=True
              )
register_agent(name="REINFORCE",
               entry_point="yarll.agents.reinforce:REINFORCEDiscreteCNN",
               state_dimensions="multi",
               action_space="discrete"
              )
register_agent(name="REINFORCE",
               entry_point="yarll.agents.reinforce:REINFORCEDiscreteCNNRNN",
               state_dimensions="multi",
               action_space="discrete",
               RNN=True
              )
register_agent(name="REINFORCE",
               entry_point="yarll.agents.reinforce:REINFORCEContinuous",
               state_dimensions="single",
               action_space="continuous"
              )
register_agent(name="SarsaFA",
               entry_point="yarll.agents.sarsa.sarsa_fa:SarsaFA",
               state_dimensions="single",
               action_space="discrete"
              )
register_agent(name="SAC",
               entry_point="yarll.agents.sac:SAC",
               state_dimensions="single",
               action_space="continuous"
               )



# -*- coding: utf8 -*-

class Policy(object):
    """Decides which action to take."""
    def __init__(self):
        super(Policy, self).__init__()

    # def select_action(self, actions, values):
    #     return



# -*- coding: utf8 -*-

from typing import List, Tuple
import numpy as np

from yarll.functionapproximation.function_approximator import FunctionApproximator

class TileCoding(FunctionApproximator):
    """Map states to tiles"""
    def __init__(self, x_low, x_high, y_low, y_high, n_tilings: int, n_y_tiles: int, n_x_tiles: int, n_actions: int) -> None:
        super(TileCoding, self).__init__(n_actions)
        self.x_low = x_low
        self.x_high = x_high
        self.y_low = y_low
        self.y_high = y_high
        self.n_x_tiles = n_x_tiles
        self.n_y_tiles = n_y_tiles
        self.n_tilings = n_tilings
        self.n_actions = n_actions

        if self.n_x_tiles % 1 != 0 or self.n_x_tiles <= 0:
            raise TypeError("Number of x tiles must be a positive natural number instead of {}".format(self.n_x_tiles))
        if self.n_y_tiles % 1 != 0 or self.n_y_tiles <= 0:
            raise TypeError("Number of y tiles must be a positive natural number instead of {}".format(self.n_y_tiles))

        self.tile_width = (self.x_high - self.x_low) / self.n_x_tiles
        self.tile_height = (self.y_high - self.y_low) / self.n_y_tiles

        self.tiling_width = self.tile_width * self.n_x_tiles
        self.tiling_height = self.tile_height * self.n_y_tiles

        self.tile_starts: List[Tuple[float, float]] = []

        # Each tiling starts at a random offset that is a fraction of the tile width and height
        for _ in range(self.n_tilings):
            self.tile_starts.append((
                self.x_low + np.random.rand() * self.tile_width,
                self.y_low + np.random.rand() * self.tile_height))

        self.features_shape = (self.n_tilings, self.n_y_tiles, self.n_x_tiles, self.n_actions)
        self.thetas = np.random.uniform(size=self.features_shape)  # Initialise randomly with values between 0 and 1

    def summed_thetas(self, state, action):
        """Theta values for features present for state and action."""
        summed = 0
        for i in range(self.n_tilings):
            shifted = state - self.tile_starts[i]  # Subtract the randomly chosen offsets
            x, y = shifted
            if (x >= 0 and x <= self.tiling_width) and (y >= 0 and y <= self.tiling_height):
                summed += self.thetas[i][int(y // self.tile_height)][int(x // self.tile_width)][action]
        return summed

    def present_features(self, state, action):
        """Features that are active for the given state and action."""
        result = np.zeros(self.thetas.shape)  # By default, all of them are inactve
        for i in range(self.n_tilings):
            shifted = state - self.tile_starts[i]
            x, y = shifted
            if(x >= 0 and x <= self.tiling_width) and(y >= 0 and y <= self.tiling_height):
                # Set the feature to active
                result[i][int(y // self.tile_height)][int(x // self.tile_width)][action] = 1
        return result



import os
import json
import argparse
import pandas as pd
import dateutil


parser = argparse.ArgumentParser()
parser.add_argument("directory", help="Path to the directory.")


def main():
    args = parser.parse_args()
    dirs = sorted([d for d in os.listdir(args.directory) if os.path.isdir(os.path.join(args.directory, d))], key=lambda x: int(x[3:]))
    header = ["RUN", "DESCR", "START", "BRANCH", "COMMITMSG"]
    data = []
    for d in dirs:
        config_path = os.path.join(args.directory, d, "config.json")
        if os.path.exists(config_path):
            with open(config_path) as f:
                config = json.load(f)
        else:
            config = {}
        run_data = [
            d,
            config.get("description", ""),
            dateutil.parser.parse(config["start_time"]).strftime("%d/%m/%y %H:%M") if "start_time" in config else "",
        ]
        run_data += [config["git"]["head"], config["git"]["message"]] if "git" in config else [""] * 2
        data.append(run_data)
    df = pd.DataFrame(data, columns=header)
    df.set_index("RUN", inplace=True)
    with pd.option_context('display.max_rows', None, 'display.max_columns', None, "display.width", None, "display.max_colwidth", 100):
        print(df)

if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-

from collections import deque
import random
import numpy as np

class Memory(object):

    def __init__(self, buffer_size: int) -> None:
        self.buffer_size: int = buffer_size
        self.num_experiences: int = 0
        self.buffer: deque = deque()

    def get_batch(self, batch_size: int) -> dict:
        # Randomly sample batch_size examples
        experiences = random.sample(self.buffer, batch_size)
        return {
            "states0": np.asarray([exp[0] for exp in experiences]),
            "actions": np.asarray([exp[1] for exp in experiences]),
            "rewards": np.asarray([exp[2] for exp in experiences]),
            "states1": np.asarray([exp[3] for exp in experiences]),
            "terminals1": np.asarray([exp[4] for exp in experiences])
        }

    def add(self, state: np.ndarray, action: np.ndarray, reward: float, new_state: np.ndarray, done: bool):
        experience = (state, action, reward, new_state, done)
        if self.num_experiences < self.buffer_size:
            self.buffer.append(experience)
            self.num_experiences += 1
        else:
            self.buffer.popleft()
            self.buffer.append(experience)

    @property
    def size(self) -> int:
        return self.buffer_size

    @property
    def n_entries(self) -> int:
        # if buffer is full, return buffer size
        # otherwise, return experience counter
        return self.num_experiences

    def erase(self):
        self.buffer = deque()
        self.num_experiences = 0



# -*- coding: utf8 -*-

import argparse
import json
import os
from os import path
import random
from typing import Any, Callable, Sequence, Union
import pkg_resources
import tensorflow as tf
from scipy import signal
import numpy as np

import gym
from gym.spaces.box import Box

def discount_rewards(x: Sequence, gamma: float) -> np.ndarray:
    """
    Given vector x, computes a vector y such that
    y[i] = x[i] + gamma * x[i+1] + gamma^2 x[i+2] + ...
    """
    return signal.lfilter([1], [1, -gamma], x[::-1], axis=0)[::-1]

# Source: http://stackoverflow.com/a/12201744/1735784
def rgb2gray(rgb: np.ndarray) -> np.ndarray:
    """
    Convert an RGB image to a grayscale image.
    Uses the formula Y' = 0.299*R + 0.587*G + 0.114*B
    """
    return np.dot(rgb[..., :3], [0.299, 0.587, 0.114])

def _process_frame42(frame: np.ndarray) -> np.ndarray:
    import cv2
    frame = frame[34:34 + 160, :160]
    # Resize by half, then down to 42x42 (essentially mipmapping). If
    # we resize directly we lose pixels that, when mapped to 42x42,
    # aren't close enough to the pixel boundary.
    frame = cv2.resize(frame, (80, 80))
    frame = cv2.resize(frame, (42, 42))
    frame = frame.mean(2)
    frame = frame.astype(np.float32)
    frame *= (1.0 / 255.0)
    frame = np.reshape(frame, [42, 42, 1])
    return frame

class AtariRescale42x42(gym.ObservationWrapper):
    def __init__(self, env=None):
        super(AtariRescale42x42, self).__init__(env)
        self.observation_space = Box(0.0, 1.0, [42, 42, 1])

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return _process_frame42(observation)


def preprocess_image(img: np.ndarray) -> np.ndarray:
    """
    Preprocess an image by converting it to grayscale and dividing its values by 256
    """
    img = img[35:195]  # crop
    img = img[::2, ::2]  # downsample by factor of 2
    return (rgb2gray(img) / 256.0)[:, :, None]

def save_config(directory: str, config: dict, envs: list):
    """Save the configuration of an agent to a file."""
    config["envs"] = envs
    # Save git information if possible
    try:
        import pygit2
        repo = pygit2.Repository(path.abspath(path.join(path.dirname(path.realpath(__file__)), "..")))
        git = {
            "head": repo.head.shorthand,
            "commit": str(repo.head.target),
            "message": repo.head.get_object().message
        }
        config["git"] = git
    except ImportError:
        pass
    with open(path.join(directory, "config.json"), "w") as outfile:
        json.dump(config, outfile, indent=4)

def json_to_dict(filename: str) -> dict:
    """Load a json file as a dictionary."""
    with open(filename) as f:
        return json.load(f)

def ge(minimum: int) -> Callable[[Any], int]:
    """Require the value for an argparse argument to be an integer >= minimum."""
    def f(value):
        ivalue = int(value)
        if ivalue < minimum:
            raise argparse.ArgumentTypeError("{} must be an integer of at least 1.".format(value))
        return ivalue
    return f

def flatten(x):
    return tf.reshape(x, [-1, np.prod(x.get_shape().as_list()[1:])])

class FastSaver(tf.train.Saver):
    """
    Disables write_meta_graph argument,
    which freezes entire process and is mostly useless.
    Source: https://github.com/openai/universe-starter-agent/blob/cbfa7901ff223adf89698f1de902811e4dabdca9/worker.py
    """
    def save(self, sess, save_path, global_step=None, latest_filename=None,
             meta_graph_suffix="meta", write_meta_graph=True):
        super(FastSaver, self).save(sess, save_path, global_step, latest_filename, meta_graph_suffix, False)

def set_seed(seed: int):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    tf.set_random_seed(seed)

def load(name: str):
    """Load an object by string."""
    entry_point = pkg_resources.EntryPoint.parse("x={}".format(name))
    result = entry_point.load(False)
    return result

def cluster_spec(num_workers: int, num_ps: int, num_masters: int = 0) -> dict:
    """
    Generate a cluster specification (for distributed Tensorflow).
    """
    cluster = {}
    port = 12222

    all_ps = []
    host = "127.0.0.1"
    for _ in range(num_ps):
        all_ps.append("{}:{}".format(host, port))
        port += 1
    cluster["ps"] = all_ps

    all_workers = []
    for _ in range(num_workers):
        all_workers.append("{}:{}".format(host, port))
        port += 1
    cluster["worker"] = all_workers

    if num_masters > 0:
        all_masters = []
        for _ in range(num_masters):
            all_masters.append("{}:{}".format(host, port))
            port += 1
        cluster["master"] = all_masters
    return cluster


class RunningMeanStd(object):
    """
    Calculates the running mean and standard deviation of values of shape `shape`.
    """

    def __init__(self, shape, epsilon=1e-2):
        super(RunningMeanStd, self).__init__()
        self.count = epsilon
        self._sum = np.zeros(shape, dtype="float64")
        self._sumsq = np.full(shape, epsilon, dtype="float64")

    def add_value(self, x):
        """
        Update count, sum and sum squared using a new value `x`.
        """
        x = np.asarray(x, dtype="float64")
        self.count += 1
        self._sum += x
        self._sumsq += np.square(x)

    def add_values(self, x):
        """
        Update count, sum and sum squared using multiple values `x`.
        """
        x = np.asarray(x, dtype="float64")
        self.count += np.shape(x)[0]
        self._sum += np.sum(x, axis=0)
        self._sumsq += np.square(x).sum(axis=0)

    @property
    def mean(self):
        return self._sum / self.count

    @property
    def std(self):
        return np.sqrt(np.maximum((self._sumsq / self.count) - np.square(self.mean), 1e-2))

number_array = Union[int, float, np.ndarray]
def normalize(x: number_array, mean: number_array, std: number_array) -> Union[float, np.ndarray]:
    if isinstance(x, np.ndarray):
        x = x.astype("float64")
    return np.clip((x - mean) / std, -5.0, 5.0)



# -*- coding: utf8 -*-

from collections import namedtuple

Experience = namedtuple("Experience", ["state", "action", "reward", "value", "features", "terminal"])

class ExperiencesMemory(object):
    """Experience gathered from an environment."""

    def __init__(self):
        super(ExperiencesMemory, self).__init__()
        self.experiences = []
        self.steps = 0

    def add(self, state, action, reward, value=None, features=None, terminal=False):
        """Add a single transition to the trajectory."""
        exp = Experience(state, action, reward, value, features, terminal)
        self.experiences.append(exp)
        self.steps += 1

    def extend(self, other):
        """
        Extend a trajectory with another one.
        """
        self.experiences.extend(other.experiences)
        self.steps += other.steps

    @property
    def states(self):
        return [exp.state for exp in self.experiences]

    @property
    def actions(self):
        return [exp.action for exp in self.experiences]

    @property
    def rewards(self):
        return [exp.reward for exp in self.experiences]

    @property
    def values(self):
        return [exp.value for exp in self.experiences]

    @property
    def features(self):
        return[exp.features for exp in self.experiences]

    @property
    def terminal(self):
        """Last experience is terminal."""
        return self.experiences[-1].terminal

    @property
    def terminals(self):
        return [exp.terminal for exp in self.experiences]



#!/usr/bin/env python
# -*- coding: utf8 -*-

import json
import argparse
import logging
from os.path import abspath, dirname
import gym

import yarll.environment
from yarll.misc.utils import ge

gym.envs.registration.logger.setLevel(logging.ERROR)

parser = argparse.ArgumentParser()
parser.add_argument("env_id", type=str, help="Name of environment for which to generate instances")
parser.add_argument("n", type=ge(1), help="Number of environments to generate")
parser.add_argument("destination", type=str, help="Where to write the environment descriptions (in json)")

def main():
    args = parser.parse_args()
    envs = yarll.environment.registration.make_random_environments(args.env_id, args.n)
    dicts = [env.metadata["parameters"] for env in envs]
    with open(args.destination, "w") as f:
        json.dump(dicts, f, indent=4)

if __name__ == '__main__':
    main()
